/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "caesar3file.h"
#include "pkwareinputstream.h"
#include "indexedimage.h"

Caesar3File::Caesar3File(const QString &filename)
	: GameFile(filename) {
	MAX_MAPSIZE = 162;
	MAX_WALKERS = 1000;
}

IndexedImage *Caesar3File::getImage() {
	// Check if file is open
	if (!in->isReadable()) {
		return NULL;
	}
	
	ok = true;
	
	// Init grids for use
	Grid<quint16> *buildings = NULL, *terrain = NULL;
	Grid<quint8> *edges = NULL, *random = NULL;
	Walker *walkers = NULL;
	int mapsize;
	char climate;
	bool is_scenario = false;

	// Check the first int. If it's zero, this is a scenario
	readInt();
	if (!readInt()) {
		is_scenario = true;
	}
	//qDebug("Start of getImage()");
	if (is_scenario) {
		in->reset();
		// mapfiles don't have building numbers, or is it the first X*X shorts?
		buildings = readShortGrid();
		edges = readByteGrid();
		terrain = readShortGrid();
		skipBytes(26244); // useless zero grid
		random = readByteGrid();
		// Mapsize and climate are easy here
		in->seek(0x335b4);
		mapsize = readInt();
		in->seek(0x33ad8);
		climate = readByte();
		//qDebug("Done reading stuff");
	} else {
		//qDebug("Saved game, reading buildings");
		buildings = readCompressedShortGrid();
		edges     = readCompressedByteGrid();
		skipCompressed(); // building IDs
		terrain   = readCompressedShortGrid();
		random     = getRandomData();
		walkers = getWalkers();
		getMapsizeAndClimate(&mapsize, &climate);
		//qDebug("Done reading stuff");
	}
	
	// Lil' sanity check
	if (!ok || mapsize > MAX_MAPSIZE) {
		//setError("Map size invalid!");
		// Clean up
		if (edges)     delete edges;
		if (terrain)   delete terrain;
		if (random)    delete random;
		if (buildings) delete buildings;
		if (walkers)   delete[] walkers;
		
		return NULL;
	}
	//qDebug("Passed map size check");
	colours = new Caesar3Colours(climate);
	IndexedImage *img = new IndexedImage(mapsize * 2, mapsize * 2);
	
	img->fill(0);
	int border = (MAX_MAPSIZE - mapsize) / 2;
	int max = border + mapsize;
	int c1, c2;
	int coords[2];
	quint8 t_random, t_edge, t_edge_below, t_edge_right;
	quint16 t_terrain, t_building;
	//qDebug("Starting terrain loop");
	for (int y = border; y < max; y++) {
		for (int x = border; x < max; x++) {
			t_terrain  = terrain->get(x, y);
			t_building = buildings->get(x, y);
			
			if (t_terrain & 0x8 && t_building != 0xc69) {
				// building & is not fort ground
				t_edge = edges->get(x, y);
				t_edge_below = edges->get(x, y+1);
				t_edge_right = edges->get(x+1, y);
				getBuildingColours(t_building, t_edge,
					t_edge_right, t_edge_below, &c1, &c2);
			} else if (!(t_terrain & 64) &&
					t_building >= 0x029c && t_building < 0x02b8) {
				// Terrain is an aquaduct *without* road beneath it
				c1 = colours->colour(Caesar3Colours::MAP_AQUA, 0);
				c2 = colours->colour(Caesar3Colours::MAP_AQUA, 1);
			} else {
				t_random = random->get(x, y);
				getTerrainColours(t_terrain, t_random, &c1, &c2);
			}
			
			// Set pixel colours
			getDiamondBitmapCoordinates(x-border, y-border, mapsize,
				&coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], c1);
			img->setIndexedPixel(coords[0]+1, coords[1], c2);
		}
	}
	//qDebug("Done doing terrain");
	delete terrain;
	delete random;
	delete buildings;
	delete edges;
	
	// Only do the walkers for saved games
	if (walkers) {
		//qDebug("Starting walkers");
		// Get walkers
		int colour;
		for (int i = 0; i < MAX_WALKERS; i++) {
			if (walkers[i].type == 0x45) { // wolf
				colour = colours->colour(Caesar3Colours::MAP_SPRITES,
					Caesar3Colours::SPRITE_WOLF);
			} else if (walkers[i].type == 0xb || walkers[i].type == 0xc ||
					walkers[i].type == 0xd) { // our soldiers
				colour = colours->colour(Caesar3Colours::MAP_SPRITES,
					Caesar3Colours::SPRITE_SOLDIER);
			} else if (walkers[i].type == 0x31 || walkers[i].type == 0x39) {
				// barbarians, rome's legions
				colour = colours->colour(Caesar3Colours::MAP_SPRITES,
					Caesar3Colours::SPRITE_BARBARIAN);
			} else if (walkers[i].type == 0x2d || walkers[i].type == 0x2f) {
				// enemies
				colour = colours->colour(Caesar3Colours::MAP_SPRITES,
					Caesar3Colours::SPRITE_ENEMY);
			} else {
				// Normal walkers don't show up
				continue;
			}
			getDiamondBitmapCoordinates(walkers[i].x, walkers[i].y, mapsize,
				&coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], colour);
			img->setIndexedPixel(coords[0]+1, coords[1], colour);
		}
		delete[] walkers;
		//qDebug("Done doing walkers");
	}
	delete colours;
	return img;
}

/**
* Figures out what the two pixel-colours for this building should be
*/
void Caesar3File::getBuildingColours(quint16 building, quint8 edge,
	quint8 edge_right, quint8 edge_below, int *c1, int *c2) {
	
	if (building >= 0xa00 && building <= 0xb06) {
		// House
		if (edge == 64) { // 1-tile house
			*c1 = colours->colour(Caesar3Colours::MAP_HOUSE, 0);
			*c2 = colours->colour(Caesar3Colours::MAP_HOUSE, 1);
		} else {
			// left pixel
			if (edge % 8 == 0 || (edge_below & 0x3f) < edge) { // left OR bottom edge
				*c1 = colours->colour(Caesar3Colours::MAP_HOUSE, 2);
			} else {
				*c1 = colours->colour(Caesar3Colours::MAP_HOUSE, 0);
			}
			// right pixel
			if (edge < 8 || edge_right != (edge & 0x3f) + 1) { // top OR right edge
				*c2 = colours->colour(Caesar3Colours::MAP_HOUSE, 3);
			} else {
				*c2 = colours->colour(Caesar3Colours::MAP_HOUSE, 1);
			}
		}
	} else if (building == 0xb2f) { // reservoir
		// left pixel
		if (edge % 8 == 0 || edge > 15) { // left edge OR bottom edge
			*c1 = colours->colour(Caesar3Colours::MAP_AQUA, 1);
		} else {
			*c1 = colours->colour(Caesar3Colours::MAP_AQUA, 0);
		}
		// right pixel
		if (edge < 8 || edge % 8 == 2) { // top edge OR right edge
			*c2 = colours->colour(Caesar3Colours::MAP_AQUA, 1);
		} else {
			*c2 = colours->colour(Caesar3Colours::MAP_AQUA, 0);
		}
	} else { // other building
		if (edge == 64) { // 1-tile building
			*c1 = colours->colour(Caesar3Colours::MAP_BUILDING, 0);
			*c2 = colours->colour(Caesar3Colours::MAP_BUILDING, 1);
		} else {
			// left pixel
			if (edge % 8 == 0 || (edge & 0x3f) + 8 != edge_below) { // left OR bottom edge
				*c1 = colours->colour(Caesar3Colours::MAP_BUILDING, 0);
			} else {
				*c1 = colours->colour(Caesar3Colours::MAP_BUILDING, 2);
			}
			// right pixel
			if (edge < 8 || edge_right != (edge & 0x3f) + 1) { // top OR right edge
				*c2 = colours->colour(Caesar3Colours::MAP_BUILDING, 1);
			} else {
				*c2 = colours->colour(Caesar3Colours::MAP_BUILDING, 3);
			}
		}
	}
}

/**
* Figures out what the terrain colours for this tile are
*/
void Caesar3File::getTerrainColours(quint16 terrain, quint8 random,
		int *c1, int *c2) {
	// 5 indicates that there is no terrain: beyond map edge
	if (terrain == 5) {
		*c1 = *c2 = colours->colour(Caesar3Colours::MAP_BACKGROUND, 0);
		return;
	}
	
	// Most terrain elements have 4 variants, depending on
	// the last 2 bits of the random number
	int num3 = random & 3;
	if (terrain & 1 || terrain & 16) { // tree/shrub
		*c1 = colours->colour(Caesar3Colours::MAP_TREE1, num3);
		*c2 = colours->colour(Caesar3Colours::MAP_TREE2, num3);
	} else if (terrain & 2 || terrain & 512) { // rock / elevation
		*c1 = colours->colour(Caesar3Colours::MAP_ROCK1, num3);
		*c2 = colours->colour(Caesar3Colours::MAP_ROCK2, num3);
	} else if (terrain & 4) { // water
		*c1 = colours->colour(Caesar3Colours::MAP_WATER1, num3);
		*c2 = colours->colour(Caesar3Colours::MAP_WATER2, num3);
	} else if (terrain & 64) { // road
		*c1 = colours->colour(Caesar3Colours::MAP_ROAD, 0);
		*c2 = colours->colour(Caesar3Colours::MAP_ROAD, 1);
	} else if (terrain & 2048) { // fertile
		*c1 = colours->colour(Caesar3Colours::MAP_FERTILE1, num3);
		*c2 = colours->colour(Caesar3Colours::MAP_FERTILE2, num3);
	} else if (terrain & 0x4000) { // wall
		*c1 = colours->colour(Caesar3Colours::MAP_WALL, 0);
		*c2 = colours->colour(Caesar3Colours::MAP_WALL, 1);
	} else { // empty land, this one has 8 variants
		*c1 = colours->colour(Caesar3Colours::MAP_EMPTY1, random & 7);
		*c2 = colours->colour(Caesar3Colours::MAP_EMPTY2, random & 7);
	}
}

/**
* Gets the random data from the saved game, which is buried after
* a few "useless" compressed chunks.
*/
Grid<quint8> *Caesar3File::getRandomData() {
	if (!ok) return NULL;
	
	// Assume the first 4 chunks of data have been read
	// The next 4 (useless) compressed parts are stored in the same way
	for (int i = 0; i < 4; i++) {
		skipCompressed();
	}
	
	// Here be the random data
	return readByteGrid();
}

/**
* Gets the walker info from the saved game, which is buried after
* a few "useless" compressed chunks.
*/
Walker * Caesar3File::getWalkers() {
	if (!ok) return NULL;
	
	// Assume random data has been read already
	// Which is followed by some more compressed data blocks
	for (int i = 0; i < 5; i++) {
		skipCompressed();
	}
	
	// Next come the walkers in compressed format
	Walker *walkers = new Walker[MAX_WALKERS];
	int length = readInt();
	
	PKWareInputStream pk(in, length, false);
	
	// Walker entries are 128 bytes
	for (int i = 0; i < MAX_WALKERS; i++) {
		if (pk.hasError()) break;
		
		pk.skip(10);
		walkers[i].type = pk.readShort();
		pk.skip(8);
		walkers[i].x = pk.readByte();
		walkers[i].y = pk.readByte();
		pk.skip(106);
	}
	pk.empty();
	
	if (pk.hasError()) {
		delete[] walkers;
		ok = false;
		return NULL;
	}
	return walkers;
}

/**
* Gets the map size and climate from the saved game, which is
* buried deep into the game file
*/
void Caesar3File::getMapsizeAndClimate(int *mapsize, char *climate) {
	if (!ok) return;
	
	// Assume walkers have been read already
	// 1200 bytes that might be uncompressed
	int size = readInt();
	if (size <= 0) {
		skipBytes(1200);
	} else {
		skipBytes(size);
	}
	
	// Two compressed blocks
	for (int i = 0; i < 2; i++) {
		skipCompressed();
	}
	
	// Three ints
	skipBytes(12);
	
	// Compressed block, followed by 70 bytes, followed by compressed
	skipCompressed();
	skipBytes(70);
	skipCompressed();
	
	// 208 bytes, followed by one compressed
	skipBytes(208);
	skipCompressed();
	
	// Start of data copied from the .map file!
	// 788 bytes to skip
	skipBytes(788);
	// Next int = map size
	*mapsize = readInt();
	
	// 1312 bytes to skip
	skipBytes(1312);
	
	// Next byte = climate
	*climate = readByte();
}

