/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "emperorfile.h"

#include "indexedimage.h"
#include "zlibfile.h"

#include <QDebug>

EmperorFile::EmperorFile(const QString &filename)
	: GameFile(filename, true) {
	MAX_MAPSIZE = 228;
	MAX_WALKERS = 2000;
	MAX_BUILDINGS = 4000;
	retrievedMaps = numMaps = 0;
}

/**
* Returns the number of maps and sets the file type
* @return The number of maps in this file. One for cities and saved games, may
*         be more for campaigns, zero for MP campaigns and unknown files
*/
int EmperorFile::getNumMaps() {
	if (!in->isOpen()) {
		return 0;
	}
	
	if (((ZlibFile*)in)->multiPart()) {
		// Scenario, find out whether it's single or multiplayer
		in->seek(6);
		if (readByte() == 0) {
			int parts = ((ZlibFile*)in)->numParts();
			if (parts <= 2) {
				filetype = TYPE_UNKNOWN;
				numMaps = 0;
			} else {
				filetype = TYPE_CAMPAIGN;
				numMaps = parts - 2;
			}
		} else {
			filetype = TYPE_MULTIPLAYER_CAMPAIGN;
			numMaps = 0;
		}
	} else {
		quint8 sig[8];
		in->read((char*)sig, 8);
		
		numMaps = 1;
		if (sig[0] == 5 && sig[1] == 0 && sig[2] == 0xfe && sig[3] == 0xca &&
			sig[4] == 0 && sig[5] == 0 && sig[6] == 2 && sig[7] == 0) {
			filetype = TYPE_MAPFILE;
		} else if (sig[0] == 0x13 && sig[1] == 0) {
			filetype = TYPE_SAVEDGAME;
		} else {
			numMaps = 0;
			filetype = TYPE_UNKNOWN;
		}
	}
	return numMaps;
}

/**
* Returns the type of the file. See the TYPE_* constants.
*/
int EmperorFile::type() const {
	return filetype;
}

IndexedImage *EmperorFile::getImage() {
	if (!in->isOpen()) {
		return NULL;
	}
	if (retrievedMaps >= numMaps) {
		return NULL;
	}
	ok = true;
	
	Grid<quint32> *terrain = NULL, *objects = NULL;
	Grid<quint8> *random = NULL, *fertile = NULL, *stone = NULL;
	Building *buildings = NULL;
	Walker *walkers = NULL;
	
	// DEBUG:
	int mapsize = MAX_MAPSIZE;
	
	if (filetype == TYPE_CAMPAIGN) {
		((ZlibFile*)in)->selectPart(retrievedMaps + 2);
	}
	
	if (filetype == TYPE_MAPFILE || filetype == TYPE_CAMPAIGN) {
		
		in->seek(84);//skipBytes(76);
		mapsize = readInt();
		skipBytes(1447 + 5 * 51984); // last byte grid = edges
		terrain = readIntGrid();
		objects = readIntGrid();
		skipBytes(51984); // byte grid
		random = readByteGrid();
		skipBytes(51984 + 8 + 2 * 51984); // byte grid + 2 ints
		fertile = readByteGrid();
		in->seek(1093607);
		buildings = getBuildings();
		
	} else if (filetype == TYPE_SAVEDGAME) {
		in->seek(10);
		
		quint8 building_pattern[8] = {'B','u','i','l','d','i','n','g'};
		if (!searchPattern(building_pattern, 8)) {
			qWarning("Building pattern not found");
		}
		// Seek 20 bytes back
		skipBytes(-20);
		//qDebug("Position before buildings: %d", (int)in->pos());
		buildings = getBuildings();
		//qDebug("Position after buildings: %d", (int)in->pos());
		walkers = getWalkers();
		//qDebug("Position after walkers: %d", (int)in->pos());
		mapsize = getMapsize(); // includes skipping weird sections
		//qDebug("Position after mapsize: %d", (int)in->pos());
		
		// seek rather than skip, version is not reliable
		quint8 pattern[16] = {0, 0, 0, 0, 5, 0, 8, 0, 5, 0, 8, 0, 5, 0, 8, 0};
		if (!searchPattern(pattern, 16)) {
			qWarning("Terrain pattern not found!");
		}
		// seek 12 bytes back.
		skipBytes(-12);
		//qDebug("Position before terrain: %d", (int)in->pos());
		terrain = readIntGrid();
		skipBytes(207936); // int grid
		//qDebug("Start of random data: %d", (int)in->pos());
		random = readByteGrid();
		skipBytes(207936); // int grid(?) with all zeroes
		skipBytes(708025); // some table
		skipBytes(51984); // real fertility grid, we need water saturation
		skipBytes(75184); // some grid plus "some" additional bytes
		stone = readByteGrid(); // resource grid, used for stone quarry depth
		skipBytes(260027); // seemingly int grid plus some extra
		//qDebug("Start of fertile data: %d", (int)in->pos());
		fertile = readByteGrid();
		skipBytes(143754);
		//qDebug("Start of object data: %d", (int)in->pos());
		objects = readIntGrid();
	}
	retrievedMaps++;
	
	if (!ok || !mapsize || mapsize > MAX_MAPSIZE) {
		//qDebug("Not ok or map size too large: %d", mapsize);
		if (terrain) delete terrain;
		if (objects) delete objects;
		if (random) delete random;
		if (stone) delete stone;
		if (fertile) delete fertile;
		if (buildings) delete[] buildings;
		if (walkers) delete[] walkers;
		return NULL;
	}
	
	// Transform it to something useful
	colours = new EmperorColours(); // all climates have the same minimap colours
	IndexedImage *img = new IndexedImage(mapsize, mapsize);
	
	int half = MAX_MAPSIZE / 2;
	int border = (MAX_MAPSIZE - mapsize) / 2;
	int max = border + mapsize;
	int c1, c2;
	int coords[2];
	int start, end;
	quint8 t_random, t_fertile, t_stone;
	quint32 t_terrain, t_object;
	//cout << "Starting loops" << endl;
	for (int y = border; y < max; y++) {
		start = (y < half) ? (border + half - y - 1) : (border + y - half);
		end   = (y < half) ? (half + y + 1 - border) : (3*half - y - border);
		for (int x = start; x < end; x++) {
			t_terrain  = terrain->get(x, y);
			t_object = objects->get(x, y);
			t_random = random->get(x, y);
			t_fertile = fertile->get(x, y);
			t_stone = (stone ? stone->get(x, y) : 255);
			
			getTerrainColours(t_terrain, t_object, t_random, t_fertile, t_stone, &c1, &c2);
			
			// Set pixel colours
			getBitmapCoordinates(x-border, y-border, mapsize, &coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], c1);
			img->setIndexedPixel(coords[0]+1, coords[1], c2);
		}
	}
	
	if (terrain) delete terrain;
	if (objects) delete objects;
	if (random) delete random;
	if (fertile) delete fertile;
	if (stone) delete stone;
	
	if (buildings) {
		placeBuildings(img, buildings, mapsize);
		delete[] buildings;
	}
	
	if (walkers) {
		int coords[2];
		int color;
		for (int i = 0; i < MAX_WALKERS; i++) {
			// Type 1 = fishing spot
			if (walkers[i].type != 0 && walkers[i].type != 1) {
				getBitmapCoordinates(walkers[i].x, walkers[i].y, mapsize,
					&coords[0], &coords[1]);
				if ((walkers[i].type & 0xff) == 0x4f) {
					color = colours->colour(EmperorColours::MAP_SPRITES,
						EmperorColours::SPRITE_HERO);
				} else if (walkers[i].type == 0x3a // cEnemyTroop
						|| walkers[i].type == 0x3b // cEnemyArcher
						|| walkers[i].type == 0x3c // cEnemyCavalry
						|| walkers[i].type == 0x3e) { // cEnemyCatapult
					color = colours->colour(EmperorColours::MAP_SPRITES,
						EmperorColours::SPRITE_ENEMY);
				} else {
					color = colours->colour(EmperorColours::MAP_SPRITES,
						EmperorColours::SPRITE_HUMAN);
				}
				img->setIndexedPixel(coords[0], coords[1], color);
				img->setIndexedPixel(coords[0]+1, coords[1], color);
			}
		}
		delete[] walkers;
	}
	
	delete colours;
	
	return img;
}

/**
* Reads the building list from a saved game or .map file
* The input should be before the start of the table, at the short before the
* integer that indicates the number of buildings in the list (1, 4000)
* @return List of the X and Y positions of buildings, with their type
*/
Building *EmperorFile::getBuildings() {
	if (!ok) return NULL;
	
	//qDebug("getBuildings(): Position now: %d", (int)in->pos());
	
	// Read header: a short of 1, and an integer with the number of buildings
	if (readShort() != 1) {
		ok = false;
		return NULL;
	}
	
	if (readInt() != (quint32)MAX_BUILDINGS) {
		ok = false;
		return NULL;
	}
	
	Building *buildings = new Building[MAX_BUILDINGS];
	quint16 type, length, version;
	
	// Go baby go!
	for (int i = 0; i < MAX_BUILDINGS && ok; i++) {
		
		type = readShort();
		if (type == 0xffff) {
			// Read type name
			if (readShort() != 0) {
				ok = false;
				break;
			}
			length = readShort();
			if (length > 255) {
				ok = false;
				break;
			}
			skipBytes(length);
		} else if (type < 10) {
			// Exception for res wall/gate in STGW-Xiangjun.map:
			// wall was removed from earlier map, but is still there in map
			// with a type of 0
			//qDebug("Position before skipping stuff: %d", (int)in->pos());
			skipBytes(128);
			while (readShort() < 0x8000) { }
			skipBytes(-2);
			//qDebug("Position after skipping stuff: %d", (int)in->pos());
			i--;
			continue;
		}
		
		version = readShort();
		buildings[i].alive = readByte();
		skipBytes(2);
		buildings[i].size = readByte();
		skipBytes(2);
		buildings[i].x = readShort();
		buildings[i].y = readShort();
		skipBytes(4);
		buildings[i].type = readShort();
		skipBytes(94);
		if (version == 5) {
			skipBytes(1);
		}
		buildings[i].rotation = readByte();
		
		// We've read 113 bytes of the record now (excluding string header)
		switch (version) {
			case 3:
				skipBytes(66); break;
			case 4:
			case 5:
				skipBytes(68); break;
			default:
				qWarning("Unknown version: %d in record %d type %d\n",
					version, i, buildings[i].type);
				ok = false;
				break;
		}
		
		// Skip the type-specific bytes here
		skipBuildingSpecificBytes(&buildings[i]);
	}
	
	if (!ok) {
		delete[] buildings;
		return NULL;
	}
	
	return buildings;
}

/**
* Skips the remaining bytes of a building record based on the type of the
* building. Sets the building colour id and may also fill in the building's
* rotation
* @param building Building of which the record should be read further
*/
void EmperorFile::skipBuildingSpecificBytes(Building *building) {
	quint16 type = building->type;
	building->colorid = 0;
	switch (type) {
		case 0:
			break; // empty slot
		
		case 0x2: // vacant house
		case 0x3: // shelter
		case 0x4: // hut
		case 0x5: // plain cottage
		case 0x6: // attractive cottage
		case 0x7: // spacious dwelling
		case 0x8: // elegant dwelling
		case 0x9: // ornate apartment
		case 0xa: // luxurious apartment
			skipBytes(65); // HouseBldg
			building->colorid = EmperorColours::MAP_HOUSING_COMMON;
			break;
		
		case 0xb: // vacant elite
		case 0xc: // abandoned elite
		case 0xd: // modest siheyuan
		case 0xe: // lavish siheyuan
		case 0xf: // humble compound
		case 0x10: // impressive compound
		case 0x11: // heavenly compound
			skipBytes(65); // HouseBldg
			building->colorid = EmperorColours::MAP_HOUSING_ELITE;
			break;
		
		case 0x1a: // cOrchard, tea bush
		case 0x1b: // cOrchard, lacquer tree
		case 0x1c: // cOrchard, mulberry tree
			skipBytes(140);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0x1f: // cFishingQuay
			skipBytes(96);
			building->rotation = readByte(); // custom rotation
			skipBytes(37);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0x21: // cHuntersTent
			skipBytes(134);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0x23: // cClayPit
		case 0x24: // cStoneQuarry
		case 0x25: // cSaltMine
		case 0x27: // cSmelter, bronze
		case 0x28: // cSmelter, iron
		case 0x29: // cSmelter, steel
		case 0x30: // cShuMint
			skipBytes(136);
			building->colorid = EmperorColours::MAP_INDUSTRY;
			break;
		
		case 0x26: // cLoggingShed
		case 0x2a: // cBronzewareMaker
		case 0x2b: // cKiln
		case 0x2c: // cLacquerwareMaker
		case 0x2d: // cPaperMaker
		case 0x2e: // cJadeCarver
		case 0x2f: // cWeaver
		case 0x31: // cMoneyPrinter
			skipBytes(134);
			building->colorid = EmperorColours::MAP_INDUSTRY;
			break;
		
		case 0x34: // cCarpentersGuild
			skipBytes(134);
			building->colorid = EmperorColours::MAP_MONUMENT_SUPPORT;
			break;
		
		case 0x35: // cMillBldg
			skipBytes(288);
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
		
		case 0x36: // cWarehouse
			skipBytes(248);
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
		
		case 0x38: // cTradingQuay
			skipBytes(96);
			building->rotation = readByte();
			skipBytes(259);
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
		
		case 0x3a: // cTradingStation
			skipBytes(348);
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
			
		case 0x3b: // cMarket, common market square
		case 0x3c: // cMarket, grand market square
			skipBytes(96);
			building->rotation = readByte();
			skipBytes(169);
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
		
		case 0x3e: // empty shop
		case 0x40: // bronzeware shop
		case 0x41: // ceramics shop
		case 0x42: // food shop
		case 0x43: // hemp shop
		case 0x44: // lacquerware shop
		case 0x45: // silk shop
		case 0x46: // tea shop
			skipBytes(140); // cStall
			building->colorid = EmperorColours::MAP_COMMERCE;
			break;
		
		case 0x47:
			skipBytes(352); // cEntertainmentSquare
			break;
		
		case 0x48:
			skipBytes(130); // cWellBldg
			building->colorid = EmperorColours::MAP_SAFETY;
			break;
		
		case 0x4b: // cTheatre
			skipBytes(236);
			building->colorid = EmperorColours::MAP_ENTERTAINMENT;
			break;
		
		case 0x4c: // tumulus
		case 0x4d: // grand tumulus
		case 0x4e: // great temple
		case 0x4f: // splendid temple
		case 0x50: // temple complex
		case 0x51: // grand temple complex
		case 0x52: // large palace
		case 0x53: // canal
		case 0x54: // underground vault
		case 0x5c: // clock tower
		case 0x5d: // grand pagoda
		case 0xfd: // earthen great wall (x4 for gate)
		case 0xfe:
		case 0xff:
		case 0x100:
		case 0x101: // stone great wall (x4 for gate)
		case 0x102:
		case 0x103:
		case 0x104:
		{
			readShort(); // cMonumentBldg
			quint16 monument_version = readShort();
			if (monument_version == 0x9) {
				skipBytes(136);
			} else if (monument_version == 0xa) {
				skipBytes(137);
			} else {
				qWarning("unknown version for monument %x: %d", type,
					monument_version);
			}
			if (type == 0x5d) {
				building->colorid = EmperorColours::MAP_MONUMENT_FILL;
			} else {
				building->colorid = EmperorColours::MAP_MONUMENT;
			}
			break;
		}
		
		case 0x59: // cResWall, brown
		case 0x5a: // cResWall, tan
		case 0x5b: // cResWall, grey
		case 0xe7: // cResWall, red
			skipBytes(132);
			building->colorid = EmperorColours::MAP_AESTHETICS;
			break;
		
		case 0x5e: // cFillBldg (for canal and wall)
			skipBytes(2);
			building->colorid = EmperorColours::MAP_MONUMENT_FILL;
			break;
		
		case 0x68: // cResGate, brown
		case 0x69: // cResGate, tan
		case 0x6a: // cResGate, grey
		case 0xe8: // cResGate, red
			skipBytes(132);
			building->colorid = EmperorColours::MAP_AESTHETICS;
			break;
		
		case 0x6e: // cPalace
			skipBytes(179);
			building->colorid = EmperorColours::MAP_GOVERNMENT;
			break;
		
		case 0x6f: // 2x2 grand way
		case 0x71: // 3x3 imperial way
			skipBytes(130); // cIndustrialBldg
			break;
		
		case 0x74: // cSculptureBldg, soldier
		case 0x75: // cSculptureBldg, qirin
		case 0x76: // cFloweringTrees, plum
		case 0x77: // cWaysidePavilion, round
		case 0x78: // cPond, pond with bridge
		case 0x79: // cTaiChiPark
		case 0x7a: // cPrivateGarden
			skipBytes(132);
			building->colorid = EmperorColours::MAP_AESTHETICS;
			break;
		
		case 0x7c: // cInspectorBldg
		case 0x7f: // cWatchtower
			skipBytes(130);
			building->colorid = EmperorColours::MAP_SAFETY;
			break;
		
		case 0x7d: // cTaxOffice
			skipBytes(132);
			building->colorid = EmperorColours::MAP_GOVERNMENT;
			break;
		
		case 0x7e: // cIndustrialBldg, roadblock
			skipBytes(130);
			building->colorid = EmperorColours::MAP_GOVERNMENT;
			break;
		
		case 0x82: // cGatehouse
			skipBytes(140);
			building->colorid = EmperorColours::MAP_MILITARY;
			break;
		
		case 0x83: // cTower
			skipBytes(132);
			building->colorid = EmperorColours::MAP_MILITARY;
			break;
		
		case 0x8a:
		case 0x97:
		case 0xb1:
		case 0xb2:
		case 0xb3:
		case 0xb4:
		case 0xb5:
		case 0xb7:
		case 0xb8: // assumed
		case 0xb9:
		case 0xba:
		case 0xbb:
		case 0xbc:
		case 0xbd:
		case 0xbe: // assumed
		case 0xbf: // assumed
		case 0xa0: // assumed
		case 0xa1:
		case 0xa2:
		case 0xa3:
		case 0xa4:
		case 0xa5:
		case 0xa6: // assumed
		case 0xa7:
		case 0xad:
		case 0xae:
		case 0xaf:
			skipBytes(130); // cIndustrialBldg, only present in .map, size = 0
			break;
		
		case 0xc0: // hemp farm
			skipBytes(142); // cSingleCropHold
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0xc1: // farmhouse
			skipBytes(222); // cMultipleCropHold
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0xc2: // cSoloField, hemp
		case 0xc3: // cFieldCrop, wheat
		case 0xc4: // cFieldCrop, millet
		case 0xc5: // cFieldCrop, rice
		case 0xc6: // cFieldCrop, cabbage
		case 0xc7: // cFieldCrop, soybean
			skipBytes(140);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0xcb: // cIrrigationPump
			skipBytes(132);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0xcf: // cHerbalistBldg
		case 0xd0: // cAcupuncturistBldg
			skipBytes(130);
			building->colorid = EmperorColours::MAP_SAFETY;
			break;
		
		case 0xd1: // cAdminCity
			skipBytes(132);
			building->colorid = EmperorColours::MAP_GOVERNMENT;
			break;
		
		case 0xd2:
			skipBytes(96); // cFerryBldg
			building->rotation = readByte(); // custom rotation
			skipBytes(2043);
			building->colorid = EmperorColours::MAP_GOVERNMENT;
			break;
		
		case 0xd3: // cMusicSchool
		case 0xd4: // cAcrobatSchool
		case 0xd5: // cDramaSchool
			skipBytes(134);
			building->colorid = EmperorColours::MAP_ENTERTAINMENT;
			break;
		
		case 0xd6: // cReligiousBldg, ancestral shrine
		case 0xd7: // cReligiousBldg, daoist shrine
		case 0xd9: // cReligiousBldg, buddhist shrine
			skipBytes(165);
			building->colorid = EmperorColours::MAP_RELIGION;
			break;
		
		case 0xd8: // cGoodReligiousBldg, daoist temple
		case 0xda: // cGoodReligiousBldg, buddist pagoda
			skipBytes(167);
			building->colorid = EmperorColours::MAP_RELIGION;
			break;
		
		case 0xdb: // cConfucianAcademy
			skipBytes(169);
			building->colorid = EmperorColours::MAP_RELIGION;
			break;
		
		case 0xdc: // cCrossbowFort
		case 0xdd: // cInfantryFort
		case 0xdf: // cSiegeFort
		case 0xe0: // cCavalryFort
		case 0xe1: // cChariotFort
			skipBytes(136);
			building->colorid = EmperorColours::MAP_MILITARY;
			break;
		
		case 0xe2: // cWeaponsmith
			skipBytes(134);
			building->colorid = EmperorColours::MAP_MILITARY;
			break;
		
		case 0xe9: // cLaborersGuild
		case 0xeb: // cMasonsGuild
		case 0xec: // cCeramistsGuild
			skipBytes(134);
			building->colorid = EmperorColours::MAP_MONUMENT_SUPPORT;
			break;
		
		case 0xed: // cOrchardHold, tea curing shed
		case 0xee: // cOrchardHold, lacquer refinery
		case 0xef: // cOrchardHold, silkworm shed
			skipBytes(142);
			building->colorid = EmperorColours::MAP_AGRICULTURE;
			break;
		
		case 0xf3: // cSculptureBldg, female
		case 0xf4: // cSculptureBldg, scholar
		case 0xf5: // cSculptureBldg, noble
		case 0xf6: // cSculptureBldg, lion
		case 0xf7: // cSculptureBldg, dragon
		case 0xf8: // cSculptureBldg, dog
		case 0xf9: // cFloweringTrees, peach
		case 0xfa: // cFloweringTrees, cherry
		case 0xfb: // cWaysidePavilion, square
		case 0xfc: // cPond, natural
			skipBytes(132);
			building->colorid = EmperorColours::MAP_AESTHETICS;
			break;
		
		default:
			qWarning("FATAL: Unknown type %x (@ position %d)", type, (int)in->pos());
			//qDebug("Type %x (x, y) = (%d, %d) size %d rotation %d", building->type,
			//	building->x, building->y, building->size, building->rotation);
			ok = false;
	}
	/*
	if (building->type) {
		qDebug("Type %x (x, y) = (%d, %d) size %d rotation %d alive %d",
			building->type, building->x, building->y, building->size,
			building->rotation, building->alive);
	}
	*/
}

Walker *EmperorFile::getWalkers() {
	if (!ok) return NULL;
	
	//qDebug("getWalkers(): Position now: %d", (int)in->pos());
	
	// Read header: a short of 1, and an integer with the number of walkers
	if (readShort() != 1) {
		ok = false;
		return NULL;
	}
	
	if (readInt() != (quint32)MAX_WALKERS) {
		ok = false;
		return NULL;
	}
	
	Walker *walkers = new Walker[MAX_WALKERS];
	quint16 type, length, version;
	
	// Go baby go!
	for (int i = 0; i < MAX_WALKERS && ok; i++) {
		
		type = readShort();
		if (type == 0xffff) {
			// Read type name
			if (readShort() != 0) {
				ok = false;
				break;
			}
			length = readShort();
			if (length > 255) {
				ok = false;
				break;
			}
			skipBytes(length);
		}
		
		version = readShort();
		if (version == 1 || version == 2) {
			skipBytes(16);
		} else {
			skipBytes(14);
		}
		walkers[i].type = readShort();
		skipBytes(11);
		walkers[i].x = readShort();
		walkers[i].y = readShort();
		
		if (version == 1 || version == 2) {
			skipBytes(362);
		} else {
			skipBytes(364);
		}
		
		// My favourite approach to variable-length records: read until you
		// hit something interesting.
		// Interesting = anything not 0, after a sequence: 0*(10*)0*
		char c = 0, prev;
		do {
			prev = c;
			in->getChar(&c);
		} while (c == 0 || c == 1 || (prev == 1 && c == 0));
		
		in->ungetChar(c);
		if (prev == 1) {
			in->ungetChar(prev);
		}
	}
	
	if (!ok) {
		delete[] walkers;
		return NULL;
	}
	
	return walkers;
}

/**
* Gets the map size for saved games
* @return The size of the map
*/
int EmperorFile::getMapsize() {
	if (!ok) return 0;
	
	// Seek until "Brief descriptio" (16 bytes)
	quint8 pattern1[16] =
		{'B','r','i','e','f',' ','d','e','s','c','r','i','p','t','i','o'};
	if (searchPattern(pattern1, 16)) {
		skipBytes(-32);
	} else {
		// Brief description not found, try another way
		quint8 pattern2[16] = 
			{255, 255, 255, 255, 0, 0, 0, 0, 0, 0, 0, 0, 255, 255, 255, 255};
		if (searchPattern(pattern2, 16)) {
			skipBytes(0x38b);
		} else {
			ok = false;
			return 0;
		}
	}
	
	// The next int is the map size
	return readInt();
}

/**
* Goes through the list of buildings and places them on the image
* @param img       Image to write on
* @param buildings List of buildings
* @param mapsize   Total size of the image
*/
void EmperorFile::placeBuildings(IndexedImage *img, Building *buildings,
		int mapsize) {
	int num;
	Building *b;
	int sizeX, sizeY;
	
	for (int i = 0; i < MAX_BUILDINGS; i++) {
		if (buildings[i].colorid == 0 || buildings[i].type == 0
			|| buildings[i].size == 0 || !buildings[i].alive) {
			continue;
		}
		// Figure out building colour ID
		b = &buildings[i];
		
		sizeX = sizeY = b->size;
		num = 0;
		
		// Special cases
		if (b->type == 0x82) { // gatehouse
			if (b->rotation) {
				sizeX = 3; sizeY = 5;
			} else {
				sizeX = 5; sizeY = 3;
			}
			for (int y = 0; y < sizeY; y++) {
				for (int x = 0; x < sizeX; x++) {
					placeBuilding(img, mapsize, b->x + x, b->y + y, 1, 1, b->colorid);
				}
			}
			continue;
		} else if (b->type == 0x1f || b->type == 0xd2) {
			// fishing quay & ferry, extra dock
			switch (b->rotation) {
				case 0:
					placeBuilding(img, mapsize, b->x, b->y - 1,
						1, 1, b->colorid);
					break;
				case 1:
					placeBuilding(img, mapsize, b->x + sizeX, b->y,
						1, 1, b->colorid);
					break;
				case 2:
					placeBuilding(img, mapsize, b->x + sizeX - 1, b->y + sizeY,
						1, 1, b->colorid);
					break;
				case 3:
					placeBuilding(img, mapsize, b->x - 1, b->y + sizeY - 1,
						1, 1, b->colorid);
					break;
			}
		} else if (b->type == 0x36 || b->type == 0x38 || b->type == 0x3a
			|| (b->colorid == EmperorColours::MAP_MONUMENT && b->type < 0xf0)) {
			// Warehouse & trading quay/post & monuments (not the wall): 1x1 buildings
			for (int y = 0; y < sizeY; y++) {
				for (int x = 0; x < sizeX; x++) {
					placeBuilding(img, mapsize, b->x + x, b->y + y, 1, 1, b->colorid);
				}
			}
			if (b->type == 0x38) {
				// add dock
				switch (b->rotation) {
					case 0:
						placeBuilding(img, mapsize, b->x + 1, b->y - 2, 2, 2, b->colorid);
						placeBuilding(img, mapsize, b->x + 1, b->y - 3, 1, 1, b->colorid);
						break;
					case 1:
						placeBuilding(img, mapsize, b->x + 4, b->y + 1, 2, 2, b->colorid);
						placeBuilding(img, mapsize, b->x + 6, b->y + 1, 1, 1, b->colorid);
						break;
					case 2:
						placeBuilding(img, mapsize, b->x + 1, b->y + 4, 2, 2, b->colorid);
						placeBuilding(img, mapsize, b->x + 2, b->y + 6, 1, 1, b->colorid);
						break;
					case 3:
						placeBuilding(img, mapsize, b->x - 2, b->y + 1, 2, 2, b->colorid);
						placeBuilding(img, mapsize, b->x - 3, b->y + 2, 1, 1, b->colorid);
						break;
				}
			}
			continue;
		} else if (b->type == 0xd1) {
			// Admin city: 2x 4x4 buildings
			placeBuilding(img, mapsize, b->x, b->y, 4, 4, b->colorid);
			if (b->rotation) {
				placeBuilding(img, mapsize, b->x + 4, b->y, 4, 4, b->colorid);
			} else {
				placeBuilding(img, mapsize, b->x, b->y + 4, 4, 4, b->colorid);
			}
			continue;
		} else if (b->type == 0x6e) {
			// Palace: 2x 5x5 buildings
			placeBuilding(img, mapsize, b->x, b->y, 5, 5, b->colorid);
			if (b->rotation) {
				placeBuilding(img, mapsize, b->x + 5, b->y, 5, 5, b->colorid);
			} else {
				placeBuilding(img, mapsize, b->x, b->y + 5, 5, 5, b->colorid);
			}
			continue;
		} else if (b->type == 0xdc || b->type == 0xdd || b->type == 0xdf
				|| b->type == 0xe0 || b->type == 0xe1) {
			// Fort, 2 2x2 buildings, plus fort ground = 4x4 in 1x1
			quint16 baseX = b->x, baseY = b->y;
			placeBuilding(img, mapsize, b->x, b->y, 2, 2, b->colorid);
			if (b->rotation) {
				placeBuilding(img, mapsize, b->x, b->y + 2, 2, 2, b->colorid);
				baseX += 2;
			} else {
				placeBuilding(img, mapsize, b->x + 2, b->y, 2, 2, b->colorid);
				baseY += 2;
			}
			for (int y = 0; y < 4; y++) {
				for (int x = 0; x < 4; x++) {
					placeBuilding(img, mapsize, baseX + x, baseY + y,
						1, 1, b->colorid);
				}
			}
			continue;
		} else if (b->type == 0x3b || b->type == 0x3c) {
			// Market squares
			if (b->rotation) {
				sizeX = 7;
				sizeY = (b->type == 0x3b) ? 4 : 6;
			} else {
				sizeY = 7;
				sizeX = (b->type == 0x3b) ? 4 : 6;
			}
			for (int y = 0; y < sizeY; y++) {
				for (int x = 0; x < sizeX; x++) {
					placeBuilding(img, mapsize, b->x + x, b->y + y,
						1, 1, b->colorid);
				}
			}
			continue;
		}
		
		// Phew, special cases over. Place the building:
		placeBuilding(img, mapsize, b->x, b->y, sizeX, sizeY, b->colorid);
	}
}

/**
* Places a building of the specified size on the specified image at the
* specified location, taking information from the edges into account
* @param img     Image to write on
* @param mapsize Total size of the image
* @param posX    X position to place building
* @param posY    Y position to place building
* @param sizeX   Width of the building
* @param sizeY   Height of the building
* @param cid     Color ID from the EmperorColours map
*/
void EmperorFile::placeBuilding(IndexedImage *img, int mapsize,
		int posX, int posY, int sizeX, int sizeY, int cid) {
	int coords[2];
	
	// Special case: 1x1 buildings (no need to get into the complex loop)
	if (sizeX == 1 && sizeY == 1) {
		getBitmapCoordinates(posX, posY, mapsize, &coords[0], &coords[1]);
		img->setIndexedPixel(coords[0], coords[1], colours->colour(cid, 0));
		if (cid == EmperorColours::MAP_MONUMENT) {
			img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 0));
		} else {
			img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 1));
		}
		return;
	}
	
	int fill = colours->colour(cid, 0);
	int top = colours->colour(cid, 1);
	int botleft = colours->colour(cid, 2);
	int botright = colours->colour(cid, 3);
	int c1, c2;
	
	if (cid == EmperorColours::MAP_COMMERCE && sizeY == 2) {
		// Exception for market stalls and other 2x2 commerce buildings
		botright = top;
	} else if (cid == EmperorColours::MAP_MONUMENT_FILL) {
		// Grand pagoda: swap two colours
		top = botleft;
		botleft = colours->colour(cid, 1);
	}
	
	for (int y = 0; y < sizeY; y++) {
		for (int x = 0; x < sizeX; x++) {
			c1 = c2 = fill;
			if (x == 0) { // top left edge
				c1 = top;
			} else if (y == sizeY - 1) { // bottom left edge
				c1 = botleft;
			}
			if (y == 0) { // top right edge
				c2 = top;
			} else if (x == sizeX - 1) { // bottom right edge
				c2 = botright;
			}
			// Exception for 2x2 industry buildings: top right corner => fill
			if (cid == EmperorColours::MAP_INDUSTRY && sizeY == 2 && !x && !y) {
				c2 = fill;
			}
			getBitmapCoordinates(posX + x, posY + y, mapsize, &coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], c1);
			img->setIndexedPixel(coords[0]+1, coords[1], c2);
		}
	}
}

/**
* Figures out what the terrain colours for this tile are
* @param terrain Terrain ID
* @param object  Object ID, used for bamboo
* @param random  Byte from the random grid
* @param fertile Fertility info
* @param stone   Stone quarry depth info
* @param c1      Out parameter: colour of the first pixel
* @param c2      Out parameter: colour of the second pixel
*/
void EmperorFile::getTerrainColours(quint32 terrain, quint32 object,
		quint8 random, quint8 fertile, quint8 stone, int *c1, int *c2) {
	
	if (terrain & 0x80000) { // "off the map"
		*c1 = *c2 = colours->colour(EmperorColours::MAP_BACKGROUND, 0);
		return;
	}
	
	int num3 = random & 3;
	if ((terrain & 0x104) == 0x100) { // flood, MUST BE FIRST
		*c1 = colours->colour(EmperorColours::MAP_WATER1, num3);
		*c2 = colours->colour(EmperorColours::MAP_WATER2, num3);
	} else if (terrain & 0x1) { // tree / bamboo
		// Check for bamboo
		if (object & 0x2) {
			*c1 = colours->colour(EmperorColours::MAP_BAMBOO1, num3);
			*c2 = colours->colour(EmperorColours::MAP_BAMBOO2, num3);
		} else {
			*c1 = colours->colour(EmperorColours::MAP_TREE1, num3);
			*c2 = colours->colour(EmperorColours::MAP_TREE2, num3);
		}
	} else if (terrain & 0x2) { // rock or ore-bearing rock
		if ((terrain & 0x300002) == 0x100002) { // copper ore
			*c1 = colours->colour(EmperorColours::MAP_COPPER1, num3);
			*c2 = colours->colour(EmperorColours::MAP_COPPER2, num3);
		} else if ((terrain & 0x300002) == 0x200002) { // iron ore
			*c1 = colours->colour(EmperorColours::MAP_IRON1, num3);
			*c2 = colours->colour(EmperorColours::MAP_IRON2, num3);
		} else { // normal rock (0x2)
			*c1 = colours->colour(EmperorColours::MAP_ROCK1, num3);
			*c2 = colours->colour(EmperorColours::MAP_ROCK2, num3);
		}
	} else if (terrain & 0x101000) { // city ruins, same as rock, MUST BE BEFORE BUILDINGS
		*c1 = colours->colour(EmperorColours::MAP_ROCK1, num3);
		*c2 = colours->colour(EmperorColours::MAP_ROCK2, num3);
	} else if (terrain & 0x10000000) { // canal / other monuments
		*c1 = *c2 = colours->colour(EmperorColours::MAP_MONUMENT, 0);
	} else if (terrain & 0x8) { // building, fill in for grand/imperial ways
		*c1 = colours->colour(EmperorColours::MAP_AESTHETICS, 0);
		*c2 = colours->colour(EmperorColours::MAP_AESTHETICS, 1);
	} else if ((terrain & 0x44) == 0x4 || (terrain & 0x104) == 0x104) {
		// water but no road, unless there's a flood
		if ((terrain & 0x4000100) == 0x4000000) { // deep water (but no flood)
			*c1 = colours->colour(EmperorColours::MAP_DEEPWATER1, num3);
			*c2 = colours->colour(EmperorColours::MAP_DEEPWATER2, num3);
		} else { // shallow water
			*c1 = colours->colour(EmperorColours::MAP_WATER1, num3);
			*c2 = colours->colour(EmperorColours::MAP_WATER2, num3);
		}
	} else if (terrain & 0x20000) { // stone quarry, stronger than elevation
		if (stone == 255) { // untouched quarry
			*c1 = colours->colour(EmperorColours::MAP_QUARRY1, num3 & 1);
			*c2 = colours->colour(EmperorColours::MAP_QUARRY2, num3 & 1);
		} else if (stone == 0x64) { // first level
			*c1 = colours->colour(EmperorColours::MAP_QUARRY1, 2 + (num3 & 1));
			*c2 = colours->colour(EmperorColours::MAP_QUARRY2, 2 + (num3 & 1));
		} else { // second level
			*c1 = colours->colour(EmperorColours::MAP_QUARRY1, 3 + (num3 & 1));
			*c2 = colours->colour(EmperorColours::MAP_QUARRY2, 3 + (num3 & 1));
		}
	} else if (terrain & 0x200) { // elevation (+ cliff, slope and whatnot)
		*c1 = colours->colour(EmperorColours::MAP_ELEVATION1, num3);
		*c2 = colours->colour(EmperorColours::MAP_ELEVATION2, num3);
	} else if (terrain & 0x20) { // garden
		*c1 = colours->colour(EmperorColours::MAP_GARDEN, 0);
		*c2 = colours->colour(EmperorColours::MAP_GARDEN, 1);
	} else if (terrain & 0x40) { // road
		*c1 = colours->colour(EmperorColours::MAP_ROAD1, num3);
		*c2 = colours->colour(EmperorColours::MAP_ROAD2, num3);
	} else if (terrain & 0x800) { // irrigation ditch
		*c1 = colours->colour(EmperorColours::MAP_IRRIGATION1, random % 3);
		*c2 = colours->colour(EmperorColours::MAP_IRRIGATION2, random % 3);
	} else if (terrain & 0x4000) { // wall
		*c1 = colours->colour(EmperorColours::MAP_WALL1, num3);
		*c2 = colours->colour(EmperorColours::MAP_WALL2, num3);
	} else if (terrain & 0x10000) { // beach
		*c1 = colours->colour(EmperorColours::MAP_BEACH1, num3);
		*c2 = colours->colour(EmperorColours::MAP_BEACH2, num3);
	} else if (terrain & 0x40000) { // salt marsh
		*c1 = colours->colour(EmperorColours::MAP_SALT1, num3);
		*c2 = colours->colour(EmperorColours::MAP_SALT2, num3);
	} else if (terrain & 0x2000000) { // pinnacle
		*c1 = colours->colour(EmperorColours::MAP_PINNACLE1, num3);
		*c2 = colours->colour(EmperorColours::MAP_PINNACLE2, num3);
	} else if (terrain & 0x80000000) {
		*c1 = colours->colour(EmperorColours::MAP_SAND1, num3);
		*c2 = colours->colour(EmperorColours::MAP_SAND2, num3);
	} else {
		if (terrain & 0x80) { // fertile land
			int num;
			if (fertile > 65) num = 0;
			else if (fertile > 55) num = 1;
			else if (fertile > 45) num = 2;
			else if (fertile > 35) num = 3;
			else if (fertile > 25) num = 4;
			else if (fertile > 15) num = 5;
			else if (fertile > 5) num = 6;
			else num = 7;
			*c1 = colours->colour(EmperorColours::MAP_FERTILE1, num);
			*c2 = colours->colour(EmperorColours::MAP_FERTILE2, num);
		} else { // empty land
			*c1 = colours->colour(EmperorColours::MAP_EMPTY1, random % 8);
			*c2 = colours->colour(EmperorColours::MAP_EMPTY2, random % 8);
		}
	}
	/*
	unsigned int mask = 0x9002000;
	if (terrain & mask) {
		// *c1 = *c2 = 0xff0000;
		qDebug("Unknown bit: %x (%x)", (terrain & mask), terrain);
	}
	*/
}
