/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "emperormapper.h"

#include "emperorthread.h"
#include "indexedimage.h"
#include "aboutdialog.h"
#include "licencedialog.h"
#include "helpdialog.h"

#include <QtGui>
#include <QMetaType>

EmperorMapper::EmperorMapper() {
	appname = QString("EmperorMapper");
	aboutDialog = NULL;
	licenceDialog = NULL;
	helpDialog = NULL;
	thread = NULL;
	filename = QDir::homePath();
	currentMap = 0;
	numMaps = 0;
	
	// Register indexedimage for thread signals
	qRegisterMetaType< QList<IndexedImage*> >("QList<IndexedImage*>");
	
	createActions();
	createMenus();
	
	imageLabel = new QLabel();
	imageLabel->setMinimumSize(226, 226);
	imageLabel->setAlignment(Qt::AlignCenter);
	
	// Show statusbar
	statusBar()->showMessage(QString(" "));
	
	setCentralWidget(imageLabel);
	setTitle();
	setWindowIcon(QIcon(":/icon.png"));
	
	// Check arguments
	QStringList args = QCoreApplication::arguments();
	if (args.size() > 1) {
		if (QFile::exists(args.at(1))) {
			loadFile(args.at(1));
		}
	}
}

EmperorMapper::~EmperorMapper() {
	for (int i = 0; i < numMaps; i++) {
		delete images[i];
	}
	if (aboutDialog) delete aboutDialog;
	if (licenceDialog) delete licenceDialog;
	if (helpDialog) delete helpDialog;
}

void EmperorMapper::open() {
	
	QString name = QFileDialog::getOpenFileName(this, tr("Open Emperor file"),
		filename,
		tr("Emperor game files (*.sav *.pak *.map);;"
		   "Emperor saved games (*.sav);;"
		   "Emperor campaigns (*.pak);;"
		   "Emperor map files (*.map)"));
	
	if (!name.isEmpty()) {
		loadFile(name);
	}
}

void EmperorMapper::loadFile(const QString &name) {
	filename = name;
	
	// Clear image & disable the save action
	imageLabel->setText(tr("Loading file..."));
	saveAction->setEnabled(false);
	for (int i = 0; i < numMaps; i++) {
		delete images[i];
		images[i] = NULL;
	}
	images.clear();
	currentMap = 0;
	
	// Create the thread, set up communication and start it
	thread = new EmperorThread(name);
	connect(thread, SIGNAL(imagesDone(const QList<IndexedImage*> &)),
			this, SLOT(handleImages(const QList<IndexedImage*>&)));
	connect(thread, SIGNAL(error(int)), this, SLOT(handleError(int)));
	connect(thread, SIGNAL(finished()), this, SLOT(threadFinished()));
	thread->start();
}

void EmperorMapper::handleImages(const QList<IndexedImage *> &images) {
	this->images = images;
	numMaps = images.size();
	
	// Set up show menu
	if (numMaps == 1) {
		showMenu->setEnabled(false);
	} else {
		// Multiple maps
		showMenu->setEnabled(true);
		for (int i = 0; i < MAX_MAPS; i++) {
			showActions[i]->setVisible(i < numMaps);
		}
		// Select parent city
		showActions[0]->setChecked(true);
	}
	
	// Set the image
	imageLabel->setPixmap(QPixmap::fromImage(*images[currentMap]));
	saveAction->setEnabled(true);
	
	setTitle(filename);
	if (numMaps == 1) {
		statusBar()->showMessage(tr("Done"));
	} else {
		statusBar()->showMessage(tr("Done. %1 maps loaded.").arg(numMaps));
	}
}

void EmperorMapper::handleError(int type) {
	numMaps = 0;
	setTitle();
	showMenu->setEnabled(false);
	
	if (type == EmperorThread::ERROR_CANTLOAD) {
		imageLabel->setText(tr("(Error loading file)"));
		QMessageBox::critical(this, tr("Error loading file"),
			tr("The file you selected could not be loaded.\n"
			   "Make sure it is a valid Emperor saved game or campaign file."));
	} else if (type == EmperorThread::ERROR_MULTIPLAYER) {
		imageLabel->setText(tr("(No maps available)"));
		QMessageBox::information(this, tr("No maps available"),
			tr("The file you selected is a multiplayer campaign.\n"
			   "No maps are stored in such campaign files."));
	}
}

void EmperorMapper::threadFinished() {
	delete thread;
	thread = NULL;
}

void EmperorMapper::save() {
	// Sanity check: can't save a null image (shouldn't occur, but still...)
	if (!images[currentMap]) {
		QMessageBox::critical(this, tr("Can't save nothing"),
			tr("There's no image to save! Load an Emperor file first."));
		return;
	}
	
	// Recommend a path to the user: filename with extension changed to png
	QString replace;
	if (numMaps == 1) {
		replace = ".png";
	} else {
		replace = QString("-%1.png").arg(currentMap+1);
	}
	
	QString recommendedPath = filename;
	int position = filename.lastIndexOf('.');
	recommendedPath.replace(position, filename.length() - position, replace);
	
	QString name = QFileDialog::getSaveFileName(this, tr("Save Image"),
		recommendedPath, tr("PNG image (*.png)"));
	
	if (name.isEmpty()) {
		return;
	}
	
	if (images[currentMap]->save(name, "png")) {
		statusBar()->showMessage(tr("Image saved"));
	} else {
		statusBar()->showMessage(tr("Failed to save image"));
	}
}

void EmperorMapper::showMap(int id) {
	if (id < 0 || id >= numMaps) {
		return;
	}
	currentMap = id;
	imageLabel->setPixmap(QPixmap::fromImage(*images[currentMap]));
}

void EmperorMapper::help() {
	if (!helpDialog) {
		QString helptext(tr(
		"EmperorMapper is a small program to create minimaps from Emperor "
		"campaigns and saved games, exactly like the in-game minimap.\n\n"
		"To load a file, use the File -> Open dialog. "
		"If you did a default installation, the location for Emperor campaigns is:\n"
		"C:\\Sierra\\EmperorRotMK\\Campaigns\n"
		"Emperor city maps are located in:\n"
		"C:\\Sierra\\EmperorRotMK\\Cities\n"
		"For Emperor saved games, the default folder is:\n"
		"C:\\Sierra\\EmperorRotMK\\Save\\YOUR_NAME\n\n"
		"Once you have selected a file, the minimap for that file will appear "
		"after a short while. "
		"For campaigns, all maps used in the campaign are loaded. If there "
		"are multiple maps available, the \"Show\" menu will be enabled so you "
		"can view and save the other maps.\n\n"
		"Using File->Save, you can save the current minimap as 256-colour PNG image."
		));
		
		helpDialog = new HelpDialog(this, appname, helptext, 305);
	}
	helpDialog->show();
}

void EmperorMapper::licence() {
	if (!licenceDialog) {
		licenceDialog = new LicenceDialog(this, appname, false);
	}
	licenceDialog->setVisible(true);
}

void EmperorMapper::about() {
	if (!aboutDialog) {
		aboutDialog = new AboutDialog(this, appname, QString("1.0 (2008-01-16)"),
			tr("Copyright (C) 2007, 2008 Bianca van Schaik &lt;pecuniam@gmail.com&gt;"),
			tr("Create minimap images from Emperor saved games and campaign files.\n\n"
			"This program is distributed in the hope that it will be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
			"See the GNU General Public License or Help->Licence for more details.\n"));
	}
	aboutDialog->exec();
}

void EmperorMapper::setTitle(const QString &filename) {
	if (filename.isEmpty()) {
		setWindowTitle(appname);
	} else {
		QFileInfo fi(filename);
		setWindowTitle(fi.fileName() + " - " + appname);
	}
}

void EmperorMapper::createActions() {
	openAction = new QAction(tr("&Open file..."), this);
	openAction->setShortcut(tr("Ctrl+O"));
	openAction->setStatusTip(tr("Open an Emperor campaign or saved game"));
	connect(openAction, SIGNAL(triggered()), this, SLOT(open()));
	
	saveAction = new QAction(tr("&Save image..."), this);
	saveAction->setShortcut(tr("Ctrl+S"));
	saveAction->setStatusTip(tr("Save minimap as PNG image"));
	saveAction->setEnabled(false);
	connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));
	
	exitAction = new QAction(tr("E&xit"), this);
	exitAction->setStatusTip(tr("Quit program"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
	
	QSignalMapper *signalMapper = new QSignalMapper(this);
	QActionGroup *actionGroup = new QActionGroup(this);
	for (int i = 0; i < MAX_MAPS; i++) {
		if (i != 9) {
			showActions[i] = new QAction(tr("Map &%1").arg(i+1), this);
		} else {
			showActions[i] = new QAction(tr("Map 1&0"), this);
		}
		showActions[i]->setStatusTip(tr("Show minimap for map %1").arg(i+1));
		showActions[i]->setCheckable(true);
		connect(showActions[i], SIGNAL(triggered()), signalMapper, SLOT(map()));
		
		signalMapper->setMapping(showActions[i], i);
		actionGroup->addAction(showActions[i]);
	}
	connect(signalMapper, SIGNAL(mapped(int)), this, SLOT(showMap(int)));
	
	helpAction = new QAction(tr("&Help"), this);
	helpAction->setShortcut(tr("F1"));
	helpAction->setStatusTip(tr("Show help file"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(help()));
	
	licenceAction = new QAction(tr("&Licence"), this);
	licenceAction->setStatusTip(tr("Show licence"));
	connect(licenceAction, SIGNAL(triggered()), this, SLOT(licence()));
	
	aboutAction = new QAction(tr("&About EmperorMapper"), this);
	aboutAction->setStatusTip(tr("About EmperorMapper"));
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
	
}

void EmperorMapper::createMenus() {
	// Set up menu
	QMenu *fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);
	
	showMenu = menuBar()->addMenu(tr("&Show"));
	for (int i = 0; i < MAX_MAPS; i++) {
		showMenu->addAction(showActions[i]);
	}
	showMenu->setEnabled(false);
	
	QMenu *helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(helpAction);
	helpMenu->addSeparator();
	helpMenu->addAction(licenceAction);
	helpMenu->addAction(aboutAction);
}

int main(int argc, char *argv[]) {
	QApplication app(argc, argv);
	EmperorMapper mainWin;
	mainWin.show();
	return app.exec();
}
