/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "pharaohfile.h"

#include "pkwareinputstream.h"
#include "indexedimage.h"

PharaohFile::PharaohFile(const QString &filename)
	: GameFile(filename) {
	MAX_MAPSIZE = 228;
	MAX_WALKERS = 2000;
	MAX_BUILDINGS = 4000;
}

IndexedImage *PharaohFile::getImage() {
	// Check if file is open
	if (!in->isReadable()) {
		return NULL;
	}
	
	ok = true;
	
	// Init grids for use
	Grid<quint32> *building_grid = NULL, *terrain = NULL;
	Grid<quint8> *edges = NULL, *random = NULL;
	Walker *walkers = NULL;
	Building *buildings = NULL;
	int mapsize;
	bool is_scenario = false;
	char fourcc[5];
	in->read(fourcc, 4);
	fourcc[4] = 0;
	if (QString(fourcc) == "MAPS") {
		is_scenario = true;
	}
	if (is_scenario) {
		// Read scenario info
		in->seek(0x177c);
		building_grid = readIntGrid();
		edges = readByteGrid();
		terrain = readIntGrid();
		skipBytes(51984);
		random = readByteGrid();
		in->seek(0x99C78);
		mapsize = readInt();
	} else {
		in->seek(0x177c);
		building_grid = readCompressedIntGrid();
		edges = readCompressedByteGrid();
		skipCompressed(); // building IDs
		terrain = readCompressedIntGrid();
		random = getRandomData();
		walkers = getWalkers();
		buildings = getBuildings();
		mapsize = getMapsize();
	}
	
	// Lil' sanity check
	if (!ok || mapsize > MAX_MAPSIZE) {
		if (building_grid) delete building_grid;
		if (terrain)       delete terrain;
		if (edges)         delete edges;
		if (random)        delete random;
		if (walkers)       delete[] walkers;
		if (buildings)     delete[] buildings;
		
		return NULL;
	}
	
	// Transform it to something useful
	colours = new PharaohColours(); // all climates have the same minimap colours
	IndexedImage *img = new IndexedImage(mapsize, mapsize);
	int half = MAX_MAPSIZE / 2;
	int border = (MAX_MAPSIZE - mapsize) / 2;
	int max = border + mapsize;
	int c1, c2;
	int coords[2];
	int start, end;
	quint8 t_random;
	quint32 t_terrain, t_building;
	
	for (int y = border; y < max; y++) {
		start = (y < half) ? (border + half - y - 1) : (border + y - half);
		end   = (y < half) ? (half + y + 1 - border) : (3*half - y - border);
		for (int x = start; x < end; x++) {
			t_terrain  = terrain->get(x, y);
			t_building = building_grid->get(x, y);
			t_random = random->get(x, y);
			getTerrainColours(t_terrain, t_random, &c1, &c2);
			if ((t_terrain & 0x48) == 0x8 && (
				(t_building >= 0x3dc6 && t_building <= 0x3ed5) ||
				(t_building >= 0x3720 && t_building <= 0x3739))) {
				// Temple complex or festival square
				c1 = colours->colour(PharaohColours::MAP_RELIGION, 0);
				c2 = colours->colour(PharaohColours::MAP_RELIGION, 1);
			}
			// Set pixel colours
			getBitmapCoordinates(x-border, y-border, mapsize, &coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], c1);
			img->setIndexedPixel(coords[0]+1, coords[1], c2);
		}
	}
	
	if (buildings) {
		placeBuildings(img, buildings, terrain, edges, mapsize);
		delete[] buildings;
	}
	delete terrain;
	delete edges;
	delete random;
	delete building_grid;
	
	if (walkers) {
		// Get walkers
		int colour;
		for (int i = 0; i < MAX_WALKERS; i++) {
			colour = walkers[i].type;
			if (walkers[i].type == 0xb || walkers[i].type == 0xc ||
					walkers[i].type == 0xd) { // our soldiers
				colour = colours->colour(PharaohColours::MAP_SPRITES,
					PharaohColours::SPRITE_SOLDIER);
			} else if (walkers[i].type == 0x14 || walkers[i].type == 0x19
				|| walkers[i].type == 0x4c // trade ship / fishing boat / ferry
				|| walkers[i].type == 0x4d || walkers[i].type == 0x4e) { // transport ship / warship
				colour = colours->colour(PharaohColours::MAP_SPRITES,
					PharaohColours::SPRITE_SHIP);
			} else if (walkers[i].type == 0x54 || walkers[i].type == 0x68) {
				// pharaoh or cleo killer animal
				colour = colours->colour(PharaohColours::MAP_SPRITES,
					PharaohColours::SPRITE_ANIMAL);
			} else if (walkers[i].type == 0x2b || walkers[i].type == 0x2c ||
				walkers[i].type == 0x2d || // enemy
				walkers[i].type == 0x36 || walkers[i].type == 0x37 || // egyption invaders
				walkers[i].type == 0x63) { // bedouins
				colour = colours->colour(PharaohColours::MAP_SPRITES,
					PharaohColours::SPRITE_ENEMY);
			} else {
				continue;
			}
			
			getBitmapCoordinates(walkers[i].x, walkers[i].y, mapsize, &coords[0], &coords[1]);
			img->setIndexedPixel(coords[0], coords[1], colour);
			img->setIndexedPixel(coords[0]+1, coords[1], colour);
		}
		delete[] walkers;
	}
	delete colours;
	
	return img;
}

/**
* Goes through the list of buildings and places them on the image
* @param img       Image to write on
* @param buildings Building list
* @param terrain   Terrain information
* @param edges     Edge information
* @param mapsize   Total size of the image
*/
void PharaohFile::placeBuildings(IndexedImage *img, Building *buildings,
		Grid<quint32> *terrain, Grid<quint8> *edges, int mapsize) {
	int cid, num;
	Building *b;
	int border = (MAX_MAPSIZE - mapsize) / 2;
	for (int i = 0; i < MAX_BUILDINGS; i++) {
		if (buildings[i].type == 0) {
			// building slot unused
			continue;
		}
		// Shorten access
		b = &buildings[i];
		
		num = 0;
		
		cid = colours->map(b->type);
		if (!cid) {
			// Unknown building
			continue;
		} else if (cid != PharaohColours::MAP_MONUMENTS
		&& (terrain->get(border + b->x, border + b->y) & 0x8) != 0x8) {
			// Building is not a building on the map
			// Most likely flooded farm
			continue;
		}
		
		// Special cases
		if (cid == PharaohColours::MAP_HOUSING && b->size > 1) {
			// house bigger than 1x1
			int coords[2];
			for (int y = 0; y < b->size; y++) {
				for (int x = 0; x < b->size; x++) {
					getBitmapCoordinates(b->x + x, b->y + y, mapsize, &coords[0], &coords[1]);
					quint8 edge = edges->get(border+b->x+x, border+b->y+y);
					quint8 edge_right = edges->get(border+b->x+x+1, border+b->y+y);
					quint8 edge_below = edges->get(border+b->x+x, border+b->y+y+1);
					if (edge % 8 == 0 || (edge_below & 0x3f) < edge) {
						img->setIndexedPixel(coords[0], coords[1], colours->colour(cid, 1));
					} else {
						img->setIndexedPixel(coords[0], coords[1], colours->colour(cid, 3));
					}
					if (edge < 8 || edge_right != (edge & 0x3f) + 1) {
						img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 0));
					} else {
						img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 2));
					}
				}
			}
			continue; // don't run place-algorithm
		} else if (b->type == 0xd1) { // festival square -- already painted
			continue;
		} else if (b->type == 0x2a) { // medium statue: all c1
			placeBuilding(img, edges, mapsize, b->x, b->y, b->size, b->size,
					colours->colour(cid, 0), colours->colour(cid, 0));
			continue;
		} else if (b->type == 0x2b) { // large statue
			int coords[2];
			for (int y = 0; y < b->size; y++) {
				for (int x = 0; x < b->size; x++) {
					getBitmapCoordinates(b->x + x, b->y + y, mapsize, &coords[0], &coords[1]);
					quint8 edge = edges->get(border+b->x+x, border+b->y+y);
					quint8 edge_right = edges->get(border+b->x+x+1, border+b->y+y);
					quint8 edge_below = edges->get(border+b->x+x, border+b->y+y+1);
					if (edge % 8 == 0 || (edge_below & 0x3f) < edge) {
						img->setIndexedPixel(coords[0], coords[1], colours->colour(cid, 2));
					} else {
						img->setIndexedPixel(coords[0], coords[1], colours->colour(cid, 0));
					}
					if (edge < 8 || edge_right != (edge & 0x3f) + 1) {
						img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 3));
					} else {
						img->setIndexedPixel(coords[0]+1, coords[1], colours->colour(cid, 0));
					}
				}
			}
			continue;
		} else if (b->type == 0xe5 || b->type == 0xea ||
				   b->type == 0xeb || b->type == 0xec) { // tombs
			int coords[2];
			int c1 = colours->colour(cid, 1);
			int c2 = colours->colour(cid, 0);
			for (int y = 0; y < b->size; y++) {
				for (int x = 0; x < b->size; x++) {
					if (!(terrain->get(border + b->x + x, border + b->y + y) & 0x40000000)) {
						getBitmapCoordinates(b->x + x, b->y + y, mapsize, &coords[0], &coords[1]);
						img->setIndexedPixel(coords[0], coords[1], c1);
						img->setIndexedPixel(coords[0]+1, coords[1], c2);
					}
				}
			}
			continue;
		} else if (b->type == 0xca) { // gatehouse
			// place a second building for the other side of the road, and
			// the road itself as well
			if (b->rotation) {
				placeBuilding(img, edges, mapsize, b->x + 3, b->y, 2, 2,
					colours->colour(cid, 0), colours->colour(cid, 1));
				placeBuilding(img, edges, mapsize, b->x + 2, b->y, 1, 2,
					colours->colour(cid, 0), colours->colour(cid, 1));
			} else {
				placeBuilding(img, edges, mapsize, b->x, b->y + 3, 2, 2,
					colours->colour(cid, 0), colours->colour(cid, 1));
				placeBuilding(img, edges, mapsize, b->x, b->y + 2, 2, 1,
					colours->colour(cid, 0), colours->colour(cid, 1));
			}
		}
		
		placeBuilding(img, edges, mapsize, b->x, b->y, b->size, b->size,
			colours->colour(cid, num), colours->colour(cid, num+1));
	}
}

/**
* Places a building of the specified size on the specified image at the
* specified location, taking information from the edges into account
* @param img     Image to write on
* @param edges   Edge information
* @param mapsize Total size of the image
* @param posX    X position to place building
* @param posY    Y position to place building
* @param sizeX   Width of the building
* @param sizeY   Height of the building
* @param c1      Colour 1 -- used for building interior
* @param c2      Colour 2 -- used for top & right edge
*/
void PharaohFile::placeBuilding(IndexedImage *img, Grid<quint8> *edges,
		int mapsize, int posX, int posY, int sizeX, int sizeY, int c1, int c2) {
	int coords[2];
	if (sizeX == 1 && sizeY == 1) {
		getBitmapCoordinates(posX, posY, mapsize, &coords[0], &coords[1]);
		img->setIndexedPixel(coords[0], coords[1], c2);
		img->setIndexedPixel(coords[0]+1, coords[1], c1);
	} else {
		int offset = (MAX_MAPSIZE - mapsize) / 2;
		for (int y = 0; y < sizeY; y++) {
			for (int x = 0; x < sizeX; x++) {
				getBitmapCoordinates(posX + x, posY + y, mapsize, &coords[0], &coords[1]);
				quint8 edge = edges->get(offset+posX+x, offset+posY+y);
				if (edge == 64) {
					img->setIndexedPixel(coords[0], coords[1], c2);
					img->setIndexedPixel(coords[0]+1, coords[1], c1);
				} else { // is not 1-tile building
					quint8 edge_right = edges->get(offset+posX+x+1, offset+posY+y);
					img->setIndexedPixel(coords[0], coords[1], c1);
					if (edge < 8 || edge_right != (edge & 0x3f) + 1 // top edge OR right edge
						|| (sizeX == 6 && x == 2) || (sizeY == 6 && y == 3)) { // 6x6 split
						img->setIndexedPixel(coords[0]+1, coords[1], c2);
					} else {
						img->setIndexedPixel(coords[0]+1, coords[1], c1);
					}
				}
			}
		}
	}
}

/**
* Figures out what the terrain colours for this tile are
*/
void PharaohFile::getTerrainColours(quint32 terrain, quint8 random, int *c1, int *c2) {
	if (terrain & 0x80000) {
		*c1 = *c2 = colours->colour(PharaohColours::MAP_BACKGROUND, 0);
		return;
	}
	
	int num3 = random & 3;
	if (terrain & 0x1) { // tree/shrub
		*c1 = colours->colour(PharaohColours::MAP_TREE1, num3);
		*c2 = colours->colour(PharaohColours::MAP_TREE2, num3);
	} else if (terrain & 0x2) { // rock
		*c1 = colours->colour(PharaohColours::MAP_ROCK1, num3);
		*c2 = colours->colour(PharaohColours::MAP_ROCK2, num3);
	} else if (terrain == 0x44) { // water + bridge = road
		*c1 = colours->colour(PharaohColours::MAP_ROAD, 0);
		*c2 = colours->colour(PharaohColours::MAP_ROAD, 1);
	} else if (terrain & 0x4) { // water
		*c1 = colours->colour(PharaohColours::MAP_WATER1, num3);
		*c2 = colours->colour(PharaohColours::MAP_WATER2, num3);
	} else if (terrain & 0x20) { // garden
		if (terrain & 0x8) { // garden + building = entertainment
			*c1 = colours->colour(PharaohColours::MAP_ENTERTAINMENT, 1);
			*c2 = colours->colour(PharaohColours::MAP_ENTERTAINMENT, 0);
		} else {
			*c1 = colours->colour(PharaohColours::MAP_AESTHETICS, 1);
			*c2 = colours->colour(PharaohColours::MAP_AESTHETICS, 0);
		}
	} else if (terrain & 0x40) { // road
		*c1 = colours->colour(PharaohColours::MAP_ROAD, 0);
		*c2 = colours->colour(PharaohColours::MAP_ROAD, 1);
	} else if (terrain & 0x100) { // irrigation
		*c1 = colours->colour(PharaohColours::MAP_WATER1, 4);
		*c2 = colours->colour(PharaohColours::MAP_WATER2, 4);
	} else if (terrain & 0x800 || terrain & 0x10000) { // meadow or floodplain
		*c1 = colours->colour(PharaohColours::MAP_FERTILE1, num3);
		*c2 = colours->colour(PharaohColours::MAP_FERTILE2, num3);
	} else if (terrain & 0x2000000) { // sand dune
		*c1 = colours->colour(PharaohColours::MAP_DUNE1, num3);
		*c2 = colours->colour(PharaohColours::MAP_DUNE2, num3);
	} else if (terrain & 0x40000) { // marshland
		*c1 = colours->colour(PharaohColours::MAP_MARSH, (random & 4) >> 2);
		*c2 = colours->colour(PharaohColours::MAP_MARSH, ((random & 4) >> 2) + 2);
	} else if (terrain & 0x804000) {// wall
		*c1 = colours->colour(PharaohColours::MAP_WALL, 0);
		*c2 = colours->colour(PharaohColours::MAP_WALL, 0);
	} else if (terrain & 0x40000000) { // tomb chamber/tunnel
		*c1 = colours->colour(PharaohColours::MAP_ROAD, 0);
		*c2 = colours->colour(PharaohColours::MAP_ROAD, 1);
	} else if (terrain & 0x10000000) { // monument plazas, etc
		*c1 = colours->colour(PharaohColours::MAP_MONUMENTS, 1);
		*c2 = colours->colour(PharaohColours::MAP_MONUMENTS, 0);
	} else { // empty land or watered land
		*c1 = colours->colour(PharaohColours::MAP_EMPTY1, random & 7);
		*c2 = colours->colour(PharaohColours::MAP_EMPTY2, random & 7);
	}
}

/**
* Reads the random data grid
*/
Grid<quint8> *PharaohFile::getRandomData() {
	// Assume the first 4 chunks of data have been read
	// The next 4 (useless) compressed parts are stored in the same way
	for (int i = 0; i < 4; i++) {
		skipCompressed();
	}
	
	// Here be the random data
	return readByteGrid();
}

/**
* Gets the walker info from the saved game, which is buried after
* a few "useless" compressed chunks.
*/
Walker * PharaohFile::getWalkers() {
	if (!ok) return NULL;
	
	// Assume random data has been read already
	// Which is followed by some more compressed data blocks
	for (int i = 0; i < 5; i++) {
		skipCompressed();
	}
	
	// Next come the walkers in compressed format
	Walker *walkers = new Walker[MAX_WALKERS];
	int length = readInt();
	
	PKWareInputStream pk(in, length, false);
	
	// Walker entry = 388 bytes
	for (int i = 0; i < MAX_WALKERS; i++) {
		if (pk.hasError()) break;
		
		pk.skip(10);
		walkers[i].type = pk.readShort();
		pk.skip(8);
		walkers[i].x = pk.readShort();
		walkers[i].y = pk.readShort();
		pk.skip(364);
	}
	pk.empty();
	
	if (pk.hasError()) {
		delete[] walkers;
		ok = false;
		return NULL;
	}
	return walkers;
}

/**
* Gets the building info from the saved game
*/
Building * PharaohFile::getBuildings() {
	if (!ok) return NULL;
	
	// Assume walkers have been read already
	
	// Three compressed blocks
	for (int i = 0; i < 3; i++) {
		skipCompressed();
	}
	
	// Three ints
	skipBytes(12);
	
	// Compressed block, followed by 72 bytes, followed by walkers
	skipCompressed();
	skipBytes(72);
	
	Building *buildings = new Building[MAX_BUILDINGS];
	int length = readInt();
	
	PKWareInputStream pk(in, length, false);
	
	// Building entry = 264 bytes
	for (int i = 0; i < MAX_BUILDINGS; i++) {
		if (pk.hasError()) break;
		
		pk.skip(3);
		buildings[i].size = pk.readByte();
		pk.skip(2);
		buildings[i].x = pk.readShort();
		buildings[i].y = pk.readShort();
		pk.skip(6);
		buildings[i].type = pk.readShort();
		pk.skip(152);
		buildings[i].rotation = pk.readByte();
		pk.skip(93);
	}
	pk.empty();
	
	if (pk.hasError()) {
		delete[] buildings;
		ok = false;
		return NULL;
	}
	return buildings;
}

/**
* Returns the map size of a saved game. Assumes everything up to
* and including buildings have been read already.
*/
int PharaohFile::getMapsize() {
	if (!ok) return 0;
	
	// Assume buildings have been read already
	
	// 68 bytes, followed by one compressed
	skipBytes(68);
	skipCompressed();
	
	// Start of data copied from the .map file!
	// 704 bytes to skip
	skipBytes(704);
	// Next int = map size
	int mapsize = readInt();
	
	return mapsize;
}
