/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "pharaohmapper.h"

#include "pharaohthread.h"
#include "indexedimage.h"
#include "aboutdialog.h"
#include "licencedialog.h"
#include "helpdialog.h"

#include <QtGui>

PharaohMapper::PharaohMapper() {
	appname = QString("PharaohMapper");
	image = NULL;
	aboutDialog = NULL;
	licenceDialog = NULL;
	helpDialog = NULL;
	filename = QDir::homePath();
	
	createActions();
	createMenus();
	
	imageLabel = new QLabel();
	imageLabel->setMinimumSize(226, 226);
	imageLabel->setAlignment(Qt::AlignCenter);
	
	// Show statusbar
	statusBar()->showMessage(QString(" "));
	
	setCentralWidget(imageLabel);
	setTitle();
	setWindowIcon(QIcon(":/icon.png"));
	
	// Check arguments
	QStringList args = QCoreApplication::arguments();
	if (args.size() > 1) {
		if (QFile::exists(args.at(1))) {
			loadFile(args.at(1));
		}
	}
}

PharaohMapper::~PharaohMapper() {
	if (image) delete image;
	if (aboutDialog) delete aboutDialog;
	if (licenceDialog) delete licenceDialog;
	if (helpDialog) delete helpDialog;
}

void PharaohMapper::open() {
	
	QString name = QFileDialog::getOpenFileName(this, tr("Open Pharaoh file"),
		filename,
		tr("Pharaoh game files (*.sav *.map);;"
		   "Pharaoh scenarios (*.map);;"
		   "Pharaoh saved games (*.sav)"));
	
	if (!name.isEmpty()) {
		loadFile(name);
	}
}

void PharaohMapper::loadFile(const QString &name) {
	filename = name;
	
	// Clear image & disable the save action
	imageLabel->setText(tr("Loading file..."));
	saveAction->setEnabled(false);
	if (image) {
		delete image;
		image = NULL;
	}
	
	// Create the thread, set up communication and start it
	thread = new PharaohThread(name);
	connect(thread, SIGNAL(imageDone(IndexedImage*)),
			this, SLOT(handleImage(IndexedImage*)));
	connect(thread, SIGNAL(error()), this, SLOT(handleError()));
	connect(thread, SIGNAL(finished()), this, SLOT(threadFinished()));
	thread->start();
}

void PharaohMapper::handleImage(IndexedImage *image) {
	this->image = image;
	
	// Set the image
	imageLabel->setPixmap(QPixmap::fromImage(*image));
	saveAction->setEnabled(true);
	
	setTitle(filename);
	statusBar()->showMessage(tr("Done"));
}

void PharaohMapper::handleError() {
	setTitle();
	
	imageLabel->setText(tr("(Error loading file)"));
	QMessageBox::critical(this, tr("Error loading file"),
		tr("The file you selected could not be loaded.\n"
		   "Make sure it is a valid Pharaoh saved game or scenario."));
}

void PharaohMapper::threadFinished() {
	delete thread;
	thread = NULL;
}

void PharaohMapper::save() {
	// Sanity check: can't save a null image (shouldn't occur, but still...)
	if (!image) {
		QMessageBox::critical(this, tr("Can't save nothing"),
			tr("There's no image to save! Load a Pharaoh file first."));
		return;
	}
	
	// Recommend a path to the user: filename with extension changed to png
	QString recommendedPath = filename;
	int position = filename.lastIndexOf('.');
	recommendedPath.replace(position, filename.length() - position, ".png");
	QString name = QFileDialog::getSaveFileName(this, tr("Save Image"),
		recommendedPath, tr("PNG image (*.png)"));
	
	if (name.isEmpty()) {
		return;
	}
	
	if (image->save(name, "png")) {
		statusBar()->showMessage(tr("Image saved"));
	} else {
		statusBar()->showMessage(tr("Failed to save image"));
	}
}

void PharaohMapper::help() {
	if (!helpDialog) {
		QString helptext(tr(
		"PharaohMapper is a small program to create minimaps from adventures "
		"and saved games, exactly like the in-game minimap. "
		"This program supports files from both Pharaoh and Cleopatra.\n\n"
		"To load a file, use the File -> Open dialog. "
		"If you did a default installation, the location for Pharaoh scenarios is:\n"
		"C:\\Sierra\\Pharaoh\\Maps\n\n"
		"For Pharaoh saved games, the default folder is:\n"
		"C:\\Sierra\\Pharaoh\\Save\\YOUR_NAME\n\n"
		"Once you have selected a file, the minimap for that file will appear "
		"after a short while.\n\n"
		"You can then save the minimap as a PNG image using File -> Save. The "
		"image is saved in 256-colour format to make the file as small as possible."
		));
		
		helpDialog = new HelpDialog(this, appname, helptext);
	}
	helpDialog->show();
}

void PharaohMapper::licence() {
	if (!licenceDialog) {
		licenceDialog = new LicenceDialog(this, appname);
	}
	licenceDialog->setVisible(true);
}

void PharaohMapper::about() {
	if (!aboutDialog) {
		aboutDialog = new AboutDialog(this, appname, QString("1.2 (2008-01-16)"),
			tr("Copyright (C) 2007, 2008 Bianca van Schaik &lt;pecuniam@gmail.com&gt;"),
			tr("Create minimap images from Pharaoh saved games and scenario files.\n\n"
			"This program is distributed in the hope that it will be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
			"See the GNU General Public License or Help->Licence for more details.\n"));
	}
	aboutDialog->exec();
}

void PharaohMapper::setTitle(const QString &filename) {
	if (filename.isEmpty()) {
		setWindowTitle(appname);
	} else {
		QFileInfo fi(filename);
		setWindowTitle(fi.fileName() + " - " + appname);
	}
}

void PharaohMapper::createActions() {
	openAction = new QAction(tr("&Open file..."), this);
	openAction->setShortcut(tr("Ctrl+O"));
	openAction->setStatusTip(tr("Open a Pharaoh scenario or saved game"));
	connect(openAction, SIGNAL(triggered()), this, SLOT(open()));
	
	saveAction = new QAction(tr("&Save image..."), this);
	saveAction->setShortcut(tr("Ctrl+S"));
	saveAction->setStatusTip(tr("Save minimap as PNG image"));
	saveAction->setEnabled(false);
	connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));
	
	exitAction = new QAction(tr("E&xit"), this);
	exitAction->setStatusTip(tr("Quit program"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
	
	helpAction = new QAction(tr("&Help"), this);
	helpAction->setShortcut(tr("F1"));
	helpAction->setStatusTip(tr("Show help file"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(help()));
	
	licenceAction = new QAction(tr("&Licence"), this);
	licenceAction->setStatusTip(tr("Show licence"));
	connect(licenceAction, SIGNAL(triggered()), this, SLOT(licence()));
	
	aboutAction = new QAction(tr("&About PharaohMapper"), this);
	aboutAction->setStatusTip(tr("About PharaohMapper"));
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
	
}

void PharaohMapper::createMenus() {
	// Set up menu
	QMenu *fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);
	
	QMenu *helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(helpAction);
	helpMenu->addSeparator();
	helpMenu->addAction(licenceAction);
	helpMenu->addAction(aboutAction);
}

int main(int argc, char *argv[]) {
	QApplication app(argc, argv);
	PharaohMapper mainWin;
	mainWin.show();
	return app.exec();
}
