/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "zeusmapper.h"

#include "zeusthread.h"
#include "indexedimage.h"
#include "aboutdialog.h"
#include "licencedialog.h"
#include "helpdialog.h"

#include <QtGui>
#include <QMetaType>

ZeusMapper::ZeusMapper() {
	appname = QString("ZeusMapper");
	aboutDialog = NULL;
	licenceDialog = NULL;
	helpDialog = NULL;
	thread = NULL;
	filename = QDir::homePath();
	currentMap = 0;
	numMaps = 0;
	
	// Register indexedimage for thread signals
	qRegisterMetaType< QList<IndexedImage*> >("QList<IndexedImage*>");
	
	createActions();
	createMenus();
	
	imageLabel = new QLabel();
	imageLabel->setMinimumSize(226, 226);
	imageLabel->setAlignment(Qt::AlignCenter);
	
	// Show statusbar
	statusBar()->showMessage(QString(" "));
	
	setCentralWidget(imageLabel);
	setTitle();
	setWindowIcon(QIcon(":/icon.png"));
	
	// Check arguments
	QStringList args = QCoreApplication::arguments();
	if (args.size() > 1) {
		if (QFile::exists(args.at(1))) {
			loadFile(args.at(1));
		}
	}
}

ZeusMapper::~ZeusMapper() {
	for (int i = 0; i < numMaps; i++) {
		delete images[i];
	}
	if (aboutDialog) delete aboutDialog;
	if (licenceDialog) delete licenceDialog;
	if (helpDialog) delete helpDialog;
}

void ZeusMapper::open() {
	
	QString name = QFileDialog::getOpenFileName(this, tr("Open Zeus file"),
		filename,
		tr("Zeus game files (*.sav *.pak *.map);;"
		   "Zeus saved games (*.sav);;Zeus adventures (*.pak);;"
		   "Zeus map files (*.map)"));
	
	if (!name.isEmpty()) {
		loadFile(name);
	}
}

void ZeusMapper::loadFile(const QString &name) {
	filename = name;
	
	// Clear image & disable the save action
	imageLabel->setText(tr("Loading file..."));
	saveAction->setEnabled(false);
	for (int i = 0; i < numMaps; i++) {
		delete images[i];
		images[i] = NULL;
	}
	images.clear();
	currentMap = 0;
	
	// Create the thread, set up communication and start it
	thread = new ZeusThread(name);
	connect(thread, SIGNAL(imagesDone(const QList<IndexedImage*> &)),
			this, SLOT(handleImages(const QList<IndexedImage*>&)));
	connect(thread, SIGNAL(error()), this, SLOT(handleError()));
	connect(thread, SIGNAL(finished()), this, SLOT(threadFinished()));
	thread->start();
}

void ZeusMapper::handleImages(const QList<IndexedImage *> &images) {
	this->images = images;
	numMaps = images.size();
	
	// Set up show menu
	if (numMaps == 1) {
		showMenu->setEnabled(false);
	} else {
		// Multiple maps
		showMenu->setEnabled(true);
		for (int i = 0; i < MAX_MAPS; i++) {
			showActions[i]->setVisible(i < numMaps);
		}
		// Select parent city
		showActions[0]->setChecked(true);
	}
	
	// Set the image
	imageLabel->setPixmap(QPixmap::fromImage(*images[currentMap]));
	saveAction->setEnabled(true);
	
	setTitle(filename);
	if (numMaps == 1) {
		statusBar()->showMessage(tr("Done"));
	} else {
		statusBar()->showMessage(tr("Done. %1 maps loaded.").arg(numMaps));
	}
}

void ZeusMapper::handleError() {
	numMaps = 0;
	setTitle();
	showMenu->setEnabled(false);
	
	imageLabel->setText(tr("(Error loading file)"));
	QMessageBox::critical(this, tr("Error loading file"),
		tr("The file you selected could not be loaded.\n"
		   "Make sure it is a valid Zeus saved game or adventure file."));
}

void ZeusMapper::threadFinished() {
	delete thread;
	thread = NULL;
}

void ZeusMapper::save() {
	// Sanity check: can't save a null image (shouldn't occur, but still...)
	if (!images[currentMap]) {
		QMessageBox::critical(this, tr("Can't save nothing"),
			tr("There's no image to save! Load a Zeus file first."));
		return;
	}
	
	// Recommend a path to the user: filename with extension changed to png
	QString replace;
	if (currentMap == 0) {
		replace = ".png";
	} else {
		replace = QString("-C%1.png").arg(currentMap);
	}
	
	QString recommendedPath = filename;
	int position = filename.lastIndexOf('.');
	recommendedPath.replace(position, filename.length() - position, replace);
	
	QString name = QFileDialog::getSaveFileName(this, tr("Save Image"),
		recommendedPath, tr("PNG image (*.png)"));
	
	if (name.isEmpty()) {
		return;
	}
	
	if (images[currentMap]->save(name, "png")) {
		statusBar()->showMessage(tr("Image saved"));
	} else {
		statusBar()->showMessage(tr("Failed to save image"));
	}
}

void ZeusMapper::showMap(int id) {
	if (id < 0 || id >= numMaps) {
		return;
	}
	currentMap = id;
	imageLabel->setPixmap(QPixmap::fromImage(*images[currentMap]));
}

void ZeusMapper::help() {
	if (!helpDialog) {
		QString helptext(tr(
		"ZeusMapper is a small program to create minimaps from adventures and "
		"saved games, exactly like the in-game minimap. "
		"This program supports files from both Zeus and Poseidon.\n\n"
		"To load a file, use the File -> Open dialog. "
		"If you did a default installation, the location for Zeus adventures is:\n"
		"C:\\Impressions Games\\Zeus\\Adventures\n"
		"For Zeus saved games, the default folder is:\n"
		"C:\\Impressions Games\\Zeus\\Save\\YOUR_NAME\n\n"
		"Once you have selected a file, the minimap for that file will appear "
		"after a short while. "
		"For adventures, both the parent city minimap and the colony maps, "
		"if any, are loaded. If there are colony maps available, the \"Show\" "
		"menu will be enabled so you can view and save the colony minimaps.\n\n"
		"Using File->Save, you can save the current minimap as 256-colour PNG image."
		));
		
		helpDialog = new HelpDialog(this, appname, helptext, 300);
	}
	helpDialog->show();
}

void ZeusMapper::licence() {
	if (!licenceDialog) {
		licenceDialog = new LicenceDialog(this, appname);
	}
	licenceDialog->setVisible(true);
}

void ZeusMapper::about() {
	if (!aboutDialog) {
		aboutDialog = new AboutDialog(this, appname, QString("1.2 (2008-01-16)"),
			tr("Copyright (C) 2007, 2008 Bianca van Schaik &lt;pecuniam@gmail.com&gt;"),
			tr("Create minimap images from Zeus saved games and adventure files.\n\n"
			"This program is distributed in the hope that it will be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
			"See the GNU General Public License or Help->Licence for more details.\n"));
	}
	aboutDialog->exec();
}

void ZeusMapper::setTitle(const QString &filename) {
	if (filename.isEmpty()) {
		setWindowTitle(appname);
	} else {
		QFileInfo fi(filename);
		setWindowTitle(fi.fileName() + " - " + appname);
	}
}

void ZeusMapper::createActions() {
	openAction = new QAction(tr("&Open file..."), this);
	openAction->setShortcut(tr("Ctrl+O"));
	openAction->setStatusTip(tr("Open a Zeus adventure or saved game"));
	connect(openAction, SIGNAL(triggered()), this, SLOT(open()));
	
	saveAction = new QAction(tr("&Save image..."), this);
	saveAction->setShortcut(tr("Ctrl+S"));
	saveAction->setStatusTip(tr("Save minimap as PNG image"));
	saveAction->setEnabled(false);
	connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));
	
	exitAction = new QAction(tr("E&xit"), this);
	exitAction->setStatusTip(tr("Quit program"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
	
	showActions[0] = new QAction(tr("&Parent city"), this);
	showActions[0]->setStatusTip(tr("Show minimap of the parent city"));
	showActions[1] = new QAction(tr("Colony &1"), this);
	showActions[1]->setStatusTip(tr("Show minimap of colony 1"));
	showActions[2] = new QAction(tr("Colony &2"), this);
	showActions[2]->setStatusTip(tr("Show minimap of colony 2"));
	showActions[3] = new QAction(tr("Colony &3"), this);
	showActions[3]->setStatusTip(tr("Show minimap of colony 3"));
	showActions[4] = new QAction(tr("Colony &4"), this);
	showActions[4]->setStatusTip(tr("Show minimap of colony 4"));
	
	QSignalMapper *signalMapper = new QSignalMapper(this);
	QActionGroup *actionGroup = new QActionGroup(this);
	for (int i = 0; i < MAX_MAPS; i++) {
		showActions[i]->setCheckable(true);
		connect(showActions[i], SIGNAL(triggered()), signalMapper, SLOT(map()));
		
		signalMapper->setMapping(showActions[i], i);
		actionGroup->addAction(showActions[i]);
	}
	connect(signalMapper, SIGNAL(mapped(int)), this, SLOT(showMap(int)));
	
	helpAction = new QAction(tr("&Help"), this);
	helpAction->setShortcut(tr("F1"));
	helpAction->setStatusTip(tr("Show help file"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(help()));
	
	licenceAction = new QAction(tr("&Licence"), this);
	licenceAction->setStatusTip(tr("Show licence"));
	connect(licenceAction, SIGNAL(triggered()), this, SLOT(licence()));
	
	aboutAction = new QAction(tr("&About ZeusMapper"), this);
	aboutAction->setStatusTip(tr("About ZeusMapper"));
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
	
}

void ZeusMapper::createMenus() {
	// Set up menu
	QMenu *fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);
	
	showMenu = menuBar()->addMenu(tr("&Show"));
	for (int i = 0; i < MAX_MAPS; i++) {
		showMenu->addAction(showActions[i]);
	}
	showMenu->setEnabled(false);
	
	QMenu *helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(helpAction);
	helpMenu->addSeparator();
	helpMenu->addAction(licenceAction);
	helpMenu->addAction(aboutAction);
}

int main(int argc, char *argv[]) {
	QApplication app(argc, argv);
	ZeusMapper mainWin;
	mainWin.show();
	return app.exec();
}
