/*
 *   Citybuilding Mappers - create minimaps from citybuilding game files
 *   Copyright (C) 2007, 2008  Bianca van Schaik
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include "zlibfile.h"
// #include <QDebug>

ZlibFile::ZlibFile(const QString &filename) {
	input = new QFile(filename);
	currentBlock = -1;
	currentPart = -1;
	maxBlocks = 0;
	filesize = 0;
	isMultiPart = false;
}

ZlibFile::~ZlibFile() {
	input->close();
	delete input;
	buffer.close();
}

qint64 ZlibFile::bytesAvailable() const {
	return QIODevice::bytesAvailable() + buffer.bytesAvailable();
}

void ZlibFile::close() {
	buffer.close();
	input->close();
}

bool ZlibFile::isSequential() const {
	return false;
}

bool ZlibFile::multiPart() const {
	return isMultiPart;
}

int ZlibFile::numParts() const {
	return maxParts;
}

bool ZlibFile::open(QIODevice::OpenMode mode) {
	if (mode & QIODevice::WriteOnly) {
		return false;
	}
	
	if (!input->open(QIODevice::ReadOnly)) {
		return false;
	}
	
	setOpenMode(mode);
	
	// Check whether it's a valid compressed file
	if (readInt() != 0xfedcbaaa) {
		return false;
	}
	
	// Check whether it's a multipart pak file
	input->seek(input->size() - 68);
	
	if (readInt() == 0xaaabcdef && readInt() == 0x3c) {
		isMultiPart = true;
		quint32 previous;
		quint32 current = 4;
		quint64 max = input->size() - 68;
		maxParts = 1;
		partPositions.append(current);
		
		for (int i = 0; i < 11; i++) {
			previous = current;
			current = readInt();
			if (previous == current || current == 0 || current >= max) {
				continue;
			}
			partPositions.append(current);
			//qDebug("Part %d at position %d", maxParts, current);
			maxParts++;
		}
		partPositions.append(max);
	} else {
		maxParts = 1;
		partPositions.append(4);
		partPositions.append(input->size());
	}
	
	// Select the first part
	return selectPart(0);
}

bool ZlibFile::selectPart(int partNumber) {
	
	if (currentPart == partNumber) {
		seek(0);
		return true;
	}
	
	if (partNumber < 0 || partNumber >= maxParts) {
		return false;
	}
	
	input->seek(partPositions[partNumber]);
	currentPart = partNumber;
	maxPosition = partPositions[partNumber + 1];
	
	prefetchBlockPositions();
	//seekToBlock(0);
	currentBlock = -1;
	
	if (readBlock(0)) {
		seek(0);
		return true;
	}
	return false;
}

bool ZlibFile::reset() {
	readBlock(0);
	return true;
}

bool ZlibFile::seek(qint64 position) {
	// Inform base class
	QIODevice::seek(position);
	
	// First determine the block we should be in
	int block = position / blocksize;
	
	// Read the block
	if (block != currentBlock) {
		if (!readBlock(block)) {
			return false;
		}
	}
	
	// Seek within the block
	if (!buffer.seek(position % blocksize)) {
		qWarning("Buffer seek failed for block %d position %d."
				 "Buffer length %d, current block %d",
			block, (int)position, (int)buffer.size(), currentBlock);
		return false;
	}
	return true;
}

qint64 ZlibFile::size() const {
	return filesize;
}

qint64 ZlibFile::readData(char *data, qint64 maxSize) {
	
	if (atEnd() || currentBlock < 0) {
		return 0;
	}
	
	qint64 available = buffer.bytesAvailable();
	qint64 bytes_read = 0;
	
	while (maxSize >= available && !atEnd()) {
		bytes_read += buffer.read(data, available);
		maxSize -= available;
		data += available;
		
		// Fetch next block
		if (!readBlock()) {
			break;
		}
		available = buffer.bytesAvailable();
	}
	
	// Read last bytes, if applicable
	if (!atEnd() && maxSize) {
		bytes_read += buffer.read(data, maxSize);
	}
	return bytes_read;
}

qint64 ZlibFile::writeData(const char * /*data*/, qint64 /*maxSize*/) {
	// Not implemented
	return 0;
}

bool ZlibFile::readBlock(int blockNumber) {
	// Determine block number
	if (blockNumber == -1) {
		blockNumber = currentBlock + 1;
	} else if (blockNumber == currentBlock) {
		return true;
	}
	
	// Seek to the block
	if (!seekToBlock(blockNumber)) {
		// Beyond end of file
		return false;
	}
	
	// input->pos() is now at the first byte of the requested block
	quint32 length = getBlockLength();
	if (length > 2 * blocksize) {
		// Something seriously wrong
		setError("Corrupted file: block length too large");
		return false;
	}
	
	QByteArray compressed = input->read(length + 4);
	if (compressed.length() < (int)length) {
		setError("Unexpected end of file");
		return false;
	}
	
	// Swap byte 0 & 3, and byte 2 & 3 (little endian -> big endian)
	char tmp = compressed[0];
	compressed[0] = compressed[3];
	compressed[3] = tmp;
	tmp = compressed[1];
	compressed[1] = compressed[2];
	compressed[2] = tmp;
	
	// Uncompress the lot
	QByteArray uncompressed = qUncompress(compressed);
	if (uncompressed.length() == 0) {
		setError(tr("Error during decompression of block %1 "
					"(length %2 uncompressed %3 position %4)")
			.arg(blockNumber)
			.arg(length)
			.arg( (compressed[0] << 24) | (compressed[1] << 16)
				| (compressed[2] << 8) | (compressed[3]) )
			.arg(input->pos())
		);
		return false;
	}
	
	// Phew, set this as new buffer & reopen the buffer
	buffer.close();
	buffer.setData(uncompressed);
	buffer.open(QIODevice::ReadOnly);
	
	currentBlock = blockNumber;
	
	return true;
}

bool ZlibFile::seekToBlock(int blockNumber) {
	if (blockNumber < 0 || blockNumber >= maxBlocks) {
		return false;
	}
	
	input->seek(blockPositions[blockNumber]);
	return true;
}

void ZlibFile::prefetchBlockPositions() {
	blockPositions.clear();
	filesize = 0;
	qint64 startPosition = input->pos();
	
	do {
		blockPositions.append(input->pos());
		quint32 length = getBlockLength();
		quint32 uncompressed_length = readInt();
		input->seek(input->pos() + length);
		maxBlocks++;
		filesize += uncompressed_length;
	} while (input->pos() < maxPosition);
	
	input->seek(startPosition);
	//qDebug("Total file length: %d, maxBlocks = %d", (int)filesize, (int)maxBlocks);
}

quint32 ZlibFile::getBlockLength() {
	// Skip first 4 bytes (check)
	input->seek(input->pos() + 4);
	return readInt();
}

quint32 ZlibFile::readInt() {
	char data[4];
	if (input->read(data, 4) != 4) {
		return 0;
	}
	
	quint32 number = 0;
	for (int i = 0; i < 4; i++) {
		number += ((unsigned char)data[i] << (i*8));
	}
	return number;
}

void ZlibFile::setError(const QString &message) {
	//qDebug() << "Error! Message: " << message;
	setErrorString(message);
}
